# coding:utf-8

"""Preferences.py - singleton preferences for CellProfiler

   TO-DO - load the default preferences from somewhere.
           Create a function to save the preferences.
           Create a function to populate a handles structure with preferences.
"""
import json
import logging
import os
import os.path
import pathlib
import sys
import tempfile
import threading
import time
import traceback
import uuid
import weakref

import psutil

from ._headless_configuration import HeadlessConfiguration
from ..constants.reader import ALL_READERS
from ..utilities.image import image_resource


LOGGER = logging.getLogger(__name__)

"""get_absolute_path - mode = output. Assume "." is the default output dir"""
ABSPATH_OUTPUT = "abspath_output"

"""get_absolute_path - mode = image. Assume "." is the default input dir"""
ABSPATH_IMAGE = "abspath_image"

__python_root = os.path.split(os.path.dirname(pathlib.Path(__file__).parent))[0]
__cp_root = os.path.split(__python_root)[0]

__is_headless = False
__headless_config = HeadlessConfiguration()


def set_headless():
    global __is_headless
    __is_headless = True


def get_headless():
    return __is_headless


__awt_headless = None


def set_awt_headless(value):
    """Tell the Javabridge to start Java with AWT headless or not

    value - True to always start Java headless, regardless of headless
            setting or other factors. False to always start Java with
            AWT enabled, despite other factors. None to use the
            default logic.

    If this is not called, Java is started with AWT headless if
    we are headless and the environment variable, CELLPROFILER_USE_XVFB,
    is not present.
    """
    global __awt_headless
    __awt_headless = value


def get_awt_headless():
    """Return True if Java is to be started with AWT headless, False to use AWT"""
    global __awt_headless
    if __awt_headless is None:
        return get_headless() and "CELLPROFILER_USE_XVFB" not in os.environ
    return __awt_headless


def get_config():
    global __is_headless, __headless_config
    if __is_headless:
        return __headless_config
    import wx

    # Return simple config on build
    if not wx.App.IsDisplayAvailable():
        return __headless_config

    if wx.App.Get() is None:
        app = wx.App(0)

    config = wx.Config.Get(False)

    if not config:
        wx.Config.Set(
            wx.Config(
                "CellProfiler",
                "BroadInstitute",
                "CellProfilerLocal.cfg",
                "CellProfilerGlobal.cfg",
                wx.CONFIG_USE_LOCAL_FILE,
            )
        )
        config = wx.Config.Get()
        if not config.Exists(PREFERENCES_VERSION):
            for key in ALL_KEYS:
                if config.Exists(key) and config.GetEntryType(key) == 1:
                    if key in BOOL_KEYS:
                        v = config.ReadBool(key)
                    elif key in INT_KEYS:
                        v = config.ReadInt(key)
                    elif key in FLOAT_KEYS:
                        v = config.ReadFloat(key)
                    else:
                        v = config.Read(key)
                    config_write(key, v)
            config_write(PREFERENCES_VERSION, str(PREFERENCES_VERSION_NUMBER))
        else:
            try:
                preferences_version_number = int(config_read(PREFERENCES_VERSION))
                if preferences_version_number != PREFERENCES_VERSION_NUMBER:
                    LOGGER.warning(
                        "Preferences version mismatch: expected %d, at %d"
                        % (PREFERENCES_VERSION_NUMBER, preferences_version_number)
                    )
            except:
                LOGGER.warning(
                    "Preferences version was %s, not a number. Resetting to current version"
                    % preferences_version_number
                )
                config_write(PREFERENCES_VERSION, str(PREFERENCES_VERSION))

    return config


def preferences_as_dict():
    # Decode settings from older CP versions which are stored as bytes.
    pref_dict = dict((k, config_read(k)) for k in ALL_KEYS)
    for key in ALL_KEYS:
        if isinstance(pref_dict[key], bytes):
            pref_dict[key] = pref_dict[key].decode("utf-8")
    # Now get the reader settings
    for reader_name, reader_class in ALL_READERS.items():
        config_stem = f"Reader.{reader_name}."
        enabled_key = config_stem + 'enabled'
        enabled_val = config_read_typed(enabled_key, bool)
        if enabled_val is not None:
            pref_dict[enabled_key] = enabled_val
        for key, _, _, key_type, default in reader_class.get_settings():
            config_key = config_stem + key
            config_val = config_read_typed(config_key, key_type)
            if config_val is not None:
                pref_dict[config_key] = config_val
    return pref_dict


def set_preferences_from_dict(d):
    """Set the preferences by faking the configuration cache"""
    global __cached_values
    __cached_values = d.copy()
    #
    # We also have to defeat value-specific caches.
    #
    global __recent_files
    __recent_files = {}
    for cache_var in (
        "__default_colormap",
        "__default_image_directory",
        "__default_output_directory",
        "__current_pipeline_path",
        "__has_reported_jvm_error",
        "__show_analysis_complete_dlg",
        "__show_exiting_test_mode_dlg",
        "__show_report_bad_sizes_dlg",
        "__show_sampling",
        "__show_workspace_choice_dlg",
        "__use_more_figure_space",
        "__warn_about_old_pipeline",
        "__workspace_file",
        "__omero_server",
        "__omero_port",
        "__omero_user",
        "__omero_session_id",
    ):
        globals()[cache_var] = None


__cached_values: dict = {}


def config_read(key):
    """Read the given configuration value

    Only read from the registry once. This is both technically efficient
    and keeps parallel running instances of CellProfiler from overwriting
    each other's values for things like the current output directory.
    """
    global __cached_values
    if not __is_headless:
        #
        # Keeps popup box from appearing during testing I hope
        #
        import wx

        shutup = wx.LogNull()
    if key in __cached_values:
        return __cached_values[key]
    if get_config().Exists(key):
        if not __is_headless:
            # Fix problems with some 32-bit
            import wx

            if key in BOOL_KEYS:
                return get_config().ReadBool(key)
            elif key in INT_KEYS:
                return get_config().ReadInt(key)
            elif key in FLOAT_KEYS:
                return get_config().ReadFloat(key)
        value = get_config().Read(key)
    else:
        value = None
    __cached_values[key] = value
    return value


def config_write(key, value):
    """Write the given configuration value
    """
    if not __is_headless:
        #
        # Keeps popup box from appearing during testing I hope
        #
        import wx

        shutup = wx.LogNull()
    __cached_values[key] = value
    if key in BOOL_KEYS:
        get_config().WriteBool(key, bool(value))
    elif key in INT_KEYS:
        get_config().WriteInt(key, int(value))
    elif key in FLOAT_KEYS:
        get_config().WriteFloat(key, float(value))
    else:
        get_config().Write(key, value)
    if not __is_headless:
        get_config().Flush()


def config_exists(key):
    """Return True if the key is defined in the configuration"""
    global __cached_values
    if key in __cached_values and __cached_values[key] is not None:
        return True
    if not get_config().Exists(key):
        return False
    # FIXME: Issue reported at https://github.com/wxWidgets/Phoenix/issues/1292, use respective .Read()
    if sys.platform != "win32":
        if get_config().GetEntryType(key) == 1:
            return get_config().Read(key) is not None
    else:
        # Bool keys
        if key in BOOL_KEYS:
            return get_config().ReadBool(key) is not None
        # Int keys
        elif key in INT_KEYS:
            return get_config().ReadInt(key) is not None
        # Double keys
        elif key in FLOAT_KEYS:
            return get_config().ReadFloat(key) is not None
        # String keys
        else:
            return get_config().Read(key) is not None
    return True


def config_read_typed(key, key_type):
    global __cached_values
    config = get_config()
    if key in __cached_values:
        return __cached_values[key]
    if config.Exists(key):
        if key_type == bool:
            value = config.ReadBool(key)
        elif key_type == int:
            value = config.ReadInt(key)
        elif key_type == float:
            value = config.ReadFloat(key)
        else:
            value = config.Read(key)
    else:
        value = None
    __cached_values[key] = value
    return value


def config_write_typed(key, value, key_type=None, flush=True):
    global __cached_values
    if key_type is None:
        # This is less safe, please do specify type.
        key_type = type(value)
    config = get_config()
    if key_type == bool:
        success = config.WriteBool(key, value)
    elif key_type == int:
        success = config.WriteInt(key, value)
    elif key_type == float:
        success = config.WriteFloat(key, value)
    else:
        success = config.Write(key, value)
    if not success:
        LOGGER.error(f"Unable to write preference key {key}")
        return
    if flush:
        config.Flush()
        __cached_values[key] = value


def export_to_json(path):
    global __is_headless
    assert not __is_headless, "wx Config can't be read in headless mode"
    config_dict = preferences_as_dict()
    # Remove recent file keys, these aren't useful.
    bad_keys = [key for key in config_dict.keys() if key.startswith('RecentFile')]
    for key in bad_keys:
        del config_dict[key]
    with open(path, mode='wt') as fd:
        json.dump(config_dict, fd)
        LOGGER.info(f"Wrote config to {path}")


def cell_profiler_root_directory():
    if __cp_root:
        return __cp_root
    return ".."


def python_root_directory():
    return __python_root


def resources_root_directory():
    if hasattr(sys, "frozen"):
        # On Mac, the application runs in CellProfiler2.0.app/Contents/Resources.
        # Not sure where this should be on PC.
        return "."
    else:
        return __python_root


DEFAULT_INPUT_FOLDER_NAME = "Default Input Folder"
DEFAULT_OUTPUT_FOLDER_NAME = "Default Output Folder"
ABSOLUTE_FOLDER_NAME = "Elsewhere..."
DEFAULT_INPUT_SUBFOLDER_NAME = "Default Input Folder sub-folder"
DEFAULT_OUTPUT_SUBFOLDER_NAME = "Default Output Folder sub-folder"
URL_FOLDER_NAME = "URL"
NO_FOLDER_NAME = "None"

"""Please add any new wordings of the above to this dictionary"""
FOLDER_CHOICE_TRANSLATIONS = {
    "Default Input Folder": DEFAULT_INPUT_FOLDER_NAME,
    "Default Output Folder": DEFAULT_OUTPUT_FOLDER_NAME,
    "Absolute path elsewhere": ABSOLUTE_FOLDER_NAME,
    "Default input directory sub-folder": DEFAULT_INPUT_SUBFOLDER_NAME,
    "Default Input Folder sub-folder": DEFAULT_INPUT_SUBFOLDER_NAME,
    "Default output directory sub-folder": DEFAULT_OUTPUT_SUBFOLDER_NAME,
    "Default Output Folder sub-folder": DEFAULT_OUTPUT_SUBFOLDER_NAME,
    "URL": URL_FOLDER_NAME,
    "None": NO_FOLDER_NAME,
    "Elsewhere...": ABSOLUTE_FOLDER_NAME,
}

PREFERENCES_VERSION = "PreferencesVersion"
PREFERENCES_VERSION_NUMBER = 1
DEFAULT_IMAGE_DIRECTORY = "DefaultImageDirectory"
DEFAULT_OUTPUT_DIRECTORY = "DefaultOutputDirectory"
TITLE_FONT_SIZE = "TitleFontSize"
TITLE_FONT_NAME = "TitleFontName"
TABLE_FONT_NAME = "TableFontName"
TABLE_FONT_SIZE = "TableFontSize"
BACKGROUND_COLOR = "BackgroundColor"
PIXEL_SIZE = "PixelSize"
COLORMAP = "Colormap"
CONSERVE_MEMORY = "ConserveMemory"
MODULEDIRECTORY = "ModuleDirectory"
SKIPVERSION = "SkipVersion2.1"
FF_RECENTFILES = "RecentFile%d"
STARTUPBLURB = "StartupBlurb"
TELEMETRY = "Telemetry"
TELEMETRY_PROMPT = "Telemetry prompt"
CHECK_FOR_UPDATES = "CheckForUpdates"
RECENT_FILE_COUNT = 10
PRIMARY_OUTLINE_COLOR = "PrimaryOutlineColor"
SECONDARY_OUTLINE_COLOR = "SecondaryOutlineColor"
TERTIARY_OUTLINE_COLOR = "TertiaryOutlineColor"
JVM_ERROR = "JVMError"
PLUGIN_DIRECTORY = "PluginDirectoryCP4"
SHOW_ANALYSIS_COMPLETE_DLG = "ShowAnalysisCompleteDlg"
SHOW_EXITING_TEST_MODE_DLG = "ShowExitingTestModeDlg"
SHOW_BAD_SIZES_DLG = "ShowBadSizesDlg"
SHOW_SAMPLING = "ShowSampling"
WARN_ABOUT_OLD_PIPELINE = "WarnAboutOldPipeline"
USE_MORE_FIGURE_SPACE = "UseMoreFigureSpace"
WORKSPACE_FILE = "WorkspaceFile"
OMERO_SERVER = "OmeroServer"
OMERO_PORT = "OmeroPort"
OMERO_USER = "OmeroUser"
OMERO_SESSION_ID = "OmeroSessionId"
MAX_WORKERS = "MaxWorkers"
TEMP_DIR = "TempDir"
WORKSPACE_CHOICE = "WorkspaceChoice"
ERROR_COLOR = "ErrorColor"
INTERPOLATION_MODE = "InterpolationMode"
INTENSITY_MODE = "IntensityMode"
NORMALIZATION_FACTOR = "NormalizationFactor"
SAVE_PIPELINE_WITH_PROJECT = "SavePipelineWithProject"
FILENAME_RE_GUESSES_FILE = "FilenameRegularExpressionGuessesFile"
PATHNAME_RE_GUESSES_FILE = "PathnameRegularExpressionGuessesFile"
CHOOSE_IMAGE_SET_FRAME_SIZE = "ChooseImageSetFrameSize"
ALWAYS_CONTINUE = "AlwaysContinue"
WIDGET_INSPECTOR = "WidgetInspector"

"""Default URL root for BatchProfiler"""

IM_NEAREST = "Nearest"
IM_BILINEAR = "Bilinear"
IM_BICUBIC = "Bicubic"

INTENSITY_MODE_RAW = "raw"
INTENSITY_MODE_NORMAL = "normalized"
INTENSITY_MODE_LOG = "log"
INTENSITY_MODE_GAMMA = "gamma"

WC_SHOW_WORKSPACE_CHOICE_DIALOG = "ShowWorkspaceChoiceDlg"
WC_OPEN_LAST_WORKSPACE = "OpenLastWorkspace"
WC_CREATE_NEW_WORKSPACE = "CreateNewWorkspace"
WC_OPEN_OLD_WORKSPACE = "OpenOldWorkspace"

"""The default extension for a CellProfiler pipeline (without the dot)"""
EXT_PIPELINE = "cppipe"

"""Possible CellProfiler pipeline extensions"""
EXT_PIPELINE_CHOICES = [EXT_PIPELINE, "cp", "cpi", "cpproj", "json"]

"""Default project extension"""
EXT_PROJECT = "cpproj"

"""Possible CellProfiler project extensions"""
EXT_PROJECT_CHOICES = [EXT_PROJECT, "cpi"]

"""Preference key for the JVM heap size in megabytes"""
JVM_HEAP_MB = "JVMHeapMB"

"""Default JVM heap size"""
DEFAULT_JVM_HEAP_MB = 512

"""Save neither the pipeline nor the file list when saving the project"""
SPP_NEITHER = "Neither"
SPP_PIPELINE_ONLY = "Pipeline"
SPP_FILE_LIST_ONLY = "File list"
SPP_PIPELINE_AND_FILE_LIST = "Pipeline and file list"
SPP_ALL = [
    SPP_NEITHER,
    SPP_PIPELINE_ONLY,
    SPP_FILE_LIST_ONLY,
    SPP_PIPELINE_AND_FILE_LIST,
]

# Registry Key Types
BOOL_KEYS = {SHOW_SAMPLING, TELEMETRY, TELEMETRY_PROMPT, STARTUPBLURB, 
             CONSERVE_MEMORY, ALWAYS_CONTINUE, WIDGET_INSPECTOR}
INT_KEYS = {SKIPVERSION, OMERO_PORT, MAX_WORKERS, JVM_HEAP_MB}
FLOAT_KEYS = {TITLE_FONT_SIZE, TABLE_FONT_SIZE, PIXEL_SIZE}

#######################
#
# Preferences help text
#
#######################
CONSERVE_MEMORY_HELP = """\
If enabled, CellProfiler will attempt to release unused system memory
after processing each image in an analysis run. Image pixel data will 
also be 'forgotten' if an image is no longer required by later modules 
in the pipeline. This can help to conserve system resources if the 
user is running other tasks in the background. Enabling this setting 
may slightly impact analysis speed, particularly during large runs.\
"""

DEFAULT_COLORMAP_HELP = """\
Specifies the color map that sets the colors for labels and other
elements. See this `page`_ for pictures of available colormaps.

.. _page: http://matplotlib.org/users/colormaps.html\
"""

DEFAULT_IMAGE_FOLDER_HELP = """\
The folder designated as the *Default Input Folder* contains the input
image or data files that you want to analyze. Several File Processing
modules (e.g., **LoadImages** or **LoadData**) provide the
option of retrieving images from this folder on a default basis unless
you specify, within the module, an alternate, specific folder on your
computer. Within modules, we recommend selecting the Default Input
Folder as much as possible, so that your pipeline will work even if you
transfer your images and pipeline to a different computer. If, instead,
you type specific folder path names into a module’s settings, your
pipeline will not work on someone else’s computer until you adjust those
pathnames within each module.

Use the *Browse* button |image0| to specify the folder you would like to
use as the Default Input Folder, or type the full folder path in the
edit box. If you type a folder path that cannot be found, the message
box below will indicate this fact until you correct the problem. If you
want to specify a folder that does not yet exist, type the desired name
and click on the *New folder* button |image1|. The folder will be
created according to the pathname you have typed.

.. |image0| image:: {BROWSE_BUTTON}
.. |image1| image:: {CREATE_BUTTON}\
""".format(
    **{
        "CREATE_BUTTON": image_resource("folder_create.png"),
        "BROWSE_BUTTON": image_resource("folder_browse.png"),
    }
)

DEFAULT_OUTPUT_FOLDER_HELP = """\
The *Default Output Folder* is accessible by pressing the “View output
settings” button at the bottom of the pipeline panel. The *Default Output
Folder* is the folder that CellProfiler uses to store the output file it
creates. Also, several File Processing modules (e.g., **SaveImages** or
**ExportToSpreadsheet**) provide the option of saving analysis results
to this folder on a default basis unless you specify, within the module,
an alternate, specific folder on your computer. Within modules, we
recommend selecting the Default Output Folder as much as possible, so
that your pipeline will work even if you transfer your images and
pipeline to a different computer. If, instead, you type specific folder
path names into a module’s settings, your pipeline will not work on
someone else’s computer until you adjust those pathnames within each
module.

Use the *Browse* button (to the right of the text box) to specify the
folder you would like to use as the Default Output Folder, or type the
full folder path in the edit box. If you type a folder path that cannot
be found, the message box below will indicate this fact until you
correct the problem. If you want to specify a folder that does not yet
exist, type the desired name and click on the *New folder* icon to the
right of the *Browse folder* icon. The folder will be created according
to the pathname you have typed.\
"""

ERROR_COLOR_HELP = "Sets the color used for the error alerts associated with misconfigured settings and other errors."

INTENSITY_MODE_HELP = """\
Sets the way CellProfiler normalizes pixel intensities when displaying.
If you choose “raw”, CellProfiler will display a pixel with a value of
“1” or above with the maximum brightness and a pixel with a value of “0”
or below as black. If you choose “normalize”, CellProfiler will find the
minimum and maximum intensities in the display image and show pixels at
maximum intensity with the maximum brightness and pixels at the minimum
intensity as black. This can be used to view dim images. If you choose
“log”, CellProfiler will use the full brightness range and will use a
log scale to scale the intensities. This can be used to view the image
background in more detail.
"""

INTERPOLATION_MODE_HELP = """\
Sets the way CellProfiler displays image pixels. If you choose
*Nearest*, CellProfiler will display each pixel as a square block of
uniform intensity. This is truest to the data, but the resulting images
look blocky and pixelated. You can choose either *Bilinear* or *Bicubic*
to see images where the a bilinear or bicubic spline model has been used
to interpolate the screen pixel value for screen pixels that do not fall
exactly in the center of the image pixel. The result, for bilinear or
bicubic interpolation is an image that is more visually appealing and
easier to interpret, but obscures the true pixel nature of the real
data.\
"""

JVM_HEAP_HELP = """\
Sets the maximum amount of memory that can be used by the Java virtual
machine. CellProfiler uses Java for loading images and for processing 
image sets. If you load extremely large images or process large image
set lists, you can use this option to start Java with a larger amount of 
memory. By default, CellProfiler starts Java with 512 MB, but you can 
override this by specifying the number of megabytes to load. You can 
also start CellProfiler from the command-line with the –jvm-heap-size
switch to get the same effect.\
"""

MAX_WORKERS_HELP = """\
Controls the maximum number of *workers* (i.e., copies of CellProfiler)
that will be started at the outset of an analysis run. CellProfiler uses
these copies to process multiple image sets in parallel, utilizing the
computer’s CPUs and memory fully. The default value is the number of
CPUs detected on your computer. Use fewer workers for pipelines that
require a large amount of memory. Use more workers for pipelines that
are accessing image data over a slow connection.

If using the **Groups** module, only one worker will be allocated to
handle each group. This means that you may have multiple workers
created, but only a subset of them may actually be active, depending on
the number of groups you have.\
"""

PLUGINS_DIRECTORY_HELP = """\
Chooses the directory that holds dynamically-loaded CellProfiler
modules. You can write your own module and place it in this directory
and CellProfiler will make it available for your pipeline. You must
restart CellProfiler after modifying this setting.\
"""

PRIMARY_OUTLINE_COLOR_HELP = """\
Sets the color used for the outline of the object of interest in the
**IdentifyPrimaryObjects**, **IdentifySecondaryObjects** and
**IdentifyTertiaryObjects** displays.\
"""

REPORT_JVM_ERROR_HELP = """\
Determines whether CellProfiler will display a warning on startup if
CellProfiler can’t locate the Java installation on your computer. Check
this box if you want to be warned. Uncheck this box to hide warnings.\
"""

SAVE_PIPELINE_WITH_PROJECT_HELP = """\
Controls whether a pipeline and/or file list file is saved whenever the
user saves the project file. Users may find it handy to have the
pipeline and/or file list saved in a readable format, for instance, for
version control whenever the project file is saved. Your project can be
restored by importing both the pipeline and file list, and your pipeline
can be run using a different file list, and your file list can be reused
by importing it into a different project. Note: When using LoadData, it
is not recommended to auto-save the file list, as this feature only
saves the file list existing in the Input Modules, not LoadData input
files.

-  *Neither:* Refrain from saving either file.
-  *Pipeline:* Save the pipeline, using the project’s file name and path
   and a .cppipe extension.
-  *File list:* Save the file list, using the project’s file name and
   path and a .txt extension.
-  *Pipeline and file list:* Save both files.\
"""

SECONDARY_OUTLINE_COLOR_HELP = """\
Sets the color used for objects other than the ones of interest. In
**IdentifyPrimaryObjects**, these are the objects that are too small or
too large. In **IdentifySecondaryObjects** and
**IdentifyTertiaryObjects**, this is the color of the secondary objects’
outline.\
"""

SHOW_ANALYSIS_COMPLETE_HELP = """\
Determines whether CellProfiler displays a message box at the end of a
run. Check this preference to show the message box or uncheck it to stop
display.\
"""

SHOW_EXITING_TEST_MODE_HELP = """\
Determines whether CellProfiler displays a message box to inform you
that a change made to the pipeline will cause test mode to end. Check
this preference to show the message box or uncheck it to stop display.\
"""

SHOW_REPORT_BAD_SIZES_DLG_HELP = """\
Determines whether CellProfiler will display a warning dialog if images
of different sizes are loaded together in an image set. Check this
preference to show the message box or uncheck it to stop display.\
"""

SHOW_SAMPLING_MENU_HELP = """\
Show the sampling menu

*Note that CellProfiler must be restarted after setting.*

The sampling menu is an interplace for Paramorama, a plugin for an
interactive visualization program for exploring the parameter space of
image analysis algorithms. will generate a text file, which specifies:
(1) all unique combinations of the sampled parameter values; (2) the
mapping from each combination of parameter values to one or more output
images; and (3) the actual output images.

More information on how to use the plugin can be found `here`_.

**References**

-  Visualization of parameter space for image analysis. Pretorius AJ,
   Bray MA, Carpenter AE and Ruddle RA. (2011) IEEE Transactions on
   Visualization and Computer Graphics, 17(12), 2402-2411.

.. _here: http://www.comp.leeds.ac.uk/scsajp/applications/paramorama2/
"""

SHOW_STARTUP_BLURB_HELP = (
    "Controls whether CellProfiler displays an orientation message on startup."
)

SHOW_TELEMETRY_HELP = """\
Allow limited and anonymous usage statistics and exception reports to be
sent to the CellProfiler team to help improve CellProfiler.\
"""

TABLE_FONT_HELP = "Sets the font used in tables displayed in module figure windows."
TERTIARY_OUTLINE_COLOR_HELP = """\
Sets the color used for the objects touching the image border or image
mask in **IdentifyPrimaryObjects**.\
"""

TEMP_DIR_HELP = """\
Sets the folder that CellProfiler uses when storing temporary files.
CellProfiler will create a temporary measurements file for analyses when
the user specifies that a MATLAB measurements file should be created or
when the user asks that no measurements file should be permanently
saved. CellProfiler will also save images accessed by http URL
temporarily to disk (but will efficiently access OMERO image planes
directly from the server).\
"""

UPDATER_HELP = """\
Allow CellProfiler to automatically check for and notify the user of new versions.
If enabled, CellProfiler will check GitHub once per week for updates.\
"""


NORMALIZATION_FACTOR_HELP = """\
Sets the normalization factor for intensity normalization methods:

-  *{INTENSITY_MODE_LOG}*: Set the gain applied to the each pixel in the image during log normalization. Pixels are
   transformed according to the formula `O = gain * log(1 + I)`, where `I` is the pixel intensity. Increasing the value
   of `gain` makes the displayed image appear brighter.
-  *{INTENSITY_MODE_GAMMA}*: Set the value of gamma. Pixels are transformed according to the formula `O = I ** gamma`,
   where I is the pixel intensity. For `gamma` > 1.0, the output image will appear darker than the original image. For
   `gamma` < 1.0, the output image will appear brighter than the original image.

The normalization factor is ignored when the normalization method is *{INTENSITY_MODE_NORMAL}* or 
*{INTENSITY_MODE_RAW}*.
""".format(
    **{
        "INTENSITY_MODE_GAMMA": INTENSITY_MODE_GAMMA,
        "INTENSITY_MODE_LOG": INTENSITY_MODE_LOG,
        "INTENSITY_MODE_NORMAL": INTENSITY_MODE_NORMAL,
        "INTENSITY_MODE_RAW": INTENSITY_MODE_RAW,
    }
)

ALWAYS_CONTINUE_HELP = """\
Determines whether CellProfiler should always skip images that 
cause errors to be raised vs stopping analysis (in headless mode)
or asking the user what to do (in GUI mode). This may help in 
processing large image sets with a small number of unusual or corrupted
images, but note that the skip of the original image set in question 
may cause issues with modules that look atmultiple image sets across 
groups (CorrectIlluminationCalculate, TrackObjects) or in Export modules,
leading to simply a later failure. Use at your own risk.
"""

WIDGET_INSPECTOR_HELP = """\
Enables wxPython Widget Inspection Tool under the "Test" menu. 
The tool displays a tree of wxWidgets and sizers in CellProfiler. 
Mostly only useful for debugging and development purposes.
"""


def recent_file(index, category=""):
    return (FF_RECENTFILES % (index + 1)) + category


"""All keys saved in the registry"""
ALL_KEYS = [
    BACKGROUND_COLOR,
    COLORMAP,
    DEFAULT_IMAGE_DIRECTORY,
    DEFAULT_OUTPUT_DIRECTORY,
    MODULEDIRECTORY,
    PLUGIN_DIRECTORY,
    PRIMARY_OUTLINE_COLOR,
    SECONDARY_OUTLINE_COLOR,
    SHOW_ANALYSIS_COMPLETE_DLG,
    SHOW_BAD_SIZES_DLG,
    SHOW_EXITING_TEST_MODE_DLG,
    WORKSPACE_CHOICE,
    SHOW_SAMPLING,
    SKIPVERSION,
    STARTUPBLURB,
    TELEMETRY,
    TELEMETRY_PROMPT,
    TABLE_FONT_NAME,
    TABLE_FONT_SIZE,
    TERTIARY_OUTLINE_COLOR,
    TITLE_FONT_NAME,
    TITLE_FONT_SIZE,
    WARN_ABOUT_OLD_PIPELINE,
    USE_MORE_FIGURE_SPACE,
    WORKSPACE_FILE,
    OMERO_SERVER,
    OMERO_PORT,
    OMERO_USER,
    SAVE_PIPELINE_WITH_PROJECT,
] + [
    recent_file(n, category)
    for n in range(RECENT_FILE_COUNT)
    for category in (
        "",
        DEFAULT_IMAGE_DIRECTORY,
        DEFAULT_OUTPUT_DIRECTORY,
        WORKSPACE_FILE,
    )
]


def module_directory():
    if not config_exists(MODULEDIRECTORY):
        return os.path.join(cell_profiler_root_directory(), "Modules")
    return str(config_read(MODULEDIRECTORY))


def set_module_directory(value):
    config_write(MODULEDIRECTORY, value)


def module_extension():
    return ".m"


__default_image_directory = None


def get_default_image_directory():
    global __default_image_directory

    if __default_image_directory is not None:
        return __default_image_directory
    # I'm not sure what it means for the preference not to exist.  No read-write preferences file?
    if not config_exists(DEFAULT_IMAGE_DIRECTORY):
        return os.path.abspath(os.path.expanduser("~"))
    # Fetch the default.  Note that it might be None
    default_image_directory = config_read(DEFAULT_IMAGE_DIRECTORY) or ""
    try:
        if os.path.isdir(default_image_directory):
            __default_image_directory = os.path.normcase(default_image_directory)
            return __default_image_directory
    except:
        LOGGER.error(
            "Unknown failure when retrieving the default image directory", exc_info=True
        )
    LOGGER.warning(
        "Warning: current path of %s is not a valid directory. Switching to home directory."
        % (default_image_directory.encode("ascii", "replace"))
    )
    # If the user's home directory is not ascii, we're not going to go hunting for one that is.
    # Fail ungracefully.
    default_image_directory = os.path.abspath(os.path.expanduser("~"))
    set_default_image_directory(default_image_directory)
    return str(os.path.normcase(default_image_directory))


def set_default_image_directory(path):
    global __default_image_directory
    __default_image_directory = path
    config_write(DEFAULT_IMAGE_DIRECTORY, path)
    add_recent_file(path, DEFAULT_IMAGE_DIRECTORY)
    fire_image_directory_changed_event()


def fire_image_directory_changed_event():
    """Notify listeners of a image directory change"""
    global __default_image_directory
    for listener in __image_directory_listeners:
        listener(PreferenceChangedEvent(__default_image_directory))


__image_directory_listeners = []


def add_image_directory_listener(listener):
    """Add a listener that will be notified when the image directory changes

    """
    __image_directory_listeners.append(listener)


def remove_image_directory_listener(listener):
    """Remove a previously-added image directory listener

    """
    if listener in __image_directory_listeners:
        __image_directory_listeners.remove(listener)


class PreferenceChangedEvent:
    def __init__(self, new_value):
        self.new_value = new_value


__default_output_directory = None


def get_default_output_directory():
    global __default_output_directory

    if __default_output_directory is not None:
        return __default_output_directory
    if not config_exists(DEFAULT_OUTPUT_DIRECTORY):
        return os.path.abspath(os.path.expanduser("~"))

    # Fetch the default.  Note that it might be None
    default_output_directory = config_read(DEFAULT_OUTPUT_DIRECTORY) or ""
    try:
        if os.path.isdir(default_output_directory):
            __default_output_directory = os.path.normcase(default_output_directory)
            return __default_output_directory
    except:
        LOGGER.error(
            "Unknown failure when retrieving the default output directory",
            exc_info=True,
        )
    LOGGER.warning(
        "Warning: current path of %s is not a valid directory. Switching to home directory."
        % (default_output_directory.encode("ascii", "replace"))
    )
    # If the user's home directory is not ascii, we're not going to go hunting for one that is.
    # Fail ungracefully.
    default_output_directory = os.path.abspath(os.path.expanduser("~"))
    set_default_output_directory(default_output_directory)
    return str(os.path.normcase(default_output_directory))


def set_default_output_directory(path):
    global __default_output_directory
    assert os.path.isdir(path), 'Default Output Folder, "%s", is not a directory' % path
    __default_output_directory = path
    config_write(DEFAULT_OUTPUT_DIRECTORY, path)
    add_recent_file(path, DEFAULT_OUTPUT_DIRECTORY)
    for listener in __output_directory_listeners:
        listener(PreferenceChangedEvent(path))


__output_directory_listeners = []


def add_output_directory_listener(listener):
    """Add a listener that will be notified when the output directory changes

    """
    __output_directory_listeners.append(listener)


def remove_output_directory_listener(listener):
    """Remove a previously-added image directory listener

    """
    if listener in __output_directory_listeners:
        __output_directory_listeners.remove(listener)


def get_title_font_size():
    if not config_exists(TITLE_FONT_SIZE):
        return 12
    title_font_size = config_read(TITLE_FONT_SIZE)
    return float(title_font_size)


def set_title_font_size(title_font_size):
    config_write(TITLE_FONT_SIZE, str(title_font_size))


def get_title_font_name():
    if not config_exists(TITLE_FONT_NAME):
        return "Tahoma"
    return config_read(TITLE_FONT_NAME)


def set_title_font_name(title_font_name):
    config_write(TITLE_FONT_NAME, title_font_name)


def get_table_font_name():
    if not config_exists(TABLE_FONT_NAME):
        return "Tahoma"
    return config_read(TABLE_FONT_NAME)


def set_table_font_name(title_font_name):
    config_write(TABLE_FONT_NAME, title_font_name)


def get_table_font_size():
    if not config_exists(TABLE_FONT_SIZE):
        return 9
    table_font_size = config_read(TABLE_FONT_SIZE)
    return float(table_font_size)


def set_table_font_size(table_font_size):
    config_write(TABLE_FONT_SIZE, str(table_font_size))


def tuple_to_color(t, default=(0, 0, 0)):
    import wx

    try:
        return wx.Colour(red=int(t[0]), green=int(t[1]), blue=int(t[2]))
    except IndexError as ValueError:
        return tuple_to_color(default)


def get_background_color():
    import wx

    return wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND)


def set_background_color(color):
    pass


__error_color = None


def get_error_color():
    """Get the color to be used for error text"""
    global __error_color
    #
    # Red found here:
    # http://www.jankoatwarpspeed.com/css-message-boxes-for-different-message-types/
    # but seems to be widely used.
    #
    default_color = (0xD8, 0x00, 0x0C)
    if __error_color is None:
        if not config_exists(ERROR_COLOR):
            __error_color = tuple_to_color(default_color)
        else:
            color_string = config_read(ERROR_COLOR)
            try:
                __error_color = tuple_to_color(color_string.split(","))
            except:
                print("Failed to parse error color string: " + color_string)
                traceback.print_exc()
                __error_color = default_color
    return __error_color


def set_error_color(color):
    """Set the color to be used for error text

    color - a WX color or ducktyped
    """
    global __error_color
    config_write(ERROR_COLOR, ",".join([str(x) for x in color.Get()]))
    __error_color = tuple_to_color(color.Get())


def get_pixel_size():
    """The size of a pixel in microns"""
    if not config_exists(PIXEL_SIZE):
        return 1.0
    return float(config_read(PIXEL_SIZE))


def set_pixel_size(pixel_size):
    config_write(PIXEL_SIZE, str(pixel_size))


def get_absolute_path(path, abspath_mode=ABSPATH_IMAGE):
    """Convert a path into an absolute path using the path conventions

    If a path starts with http:, https: or ftp:, leave it unchanged.
    If a path starts with "./", then make the path relative to the
    Default Output Folder.
    If a path starts with "&/", then make the path relative to the
    Default Input Folder.
    If a "path" has no path component then make the path relative to
    the Default Output Folder.
    """

    if abspath_mode == ABSPATH_OUTPUT:
        osep = "."
        isep = "&"
    elif abspath_mode == ABSPATH_IMAGE:
        osep = "&"
        isep = "."
    else:
        raise ValueError("Unknown abspath mode: %s" % abspath_mode)
    if is_url_path(path):
        return path
    if path.startswith(osep + os.path.sep) or (
        "altsep" in os.path.__all__
        and os.path.altsep
        and path.startswith(osep + os.path.altsep)
    ):
        return os.path.join(get_default_output_directory(), path[2:])
    elif path.startswith(isep + os.path.sep) or (
        "altsep" in os.path.__all__
        and os.path.altsep
        and path.startswith(isep + os.path.altsep)
    ):
        return os.path.join(get_default_image_directory(), path[2:])
    elif len(os.path.split(path)[0]) == 0:
        return os.path.join(get_default_output_directory(), path)
    else:
        return str(os.path.normpath(os.path.abspath(path)))


def is_url_path(path):
    """Return True if the path should be treated as a URL"""
    for protocol in ("http", "https", "ftp", "s3", "gs"):
        if path.lower().startswith("%s:" % protocol):
            return True
    return False


__default_colormap = None


def get_default_colormap():
    global __default_colormap
    if __default_colormap is None:
        if not config_exists(COLORMAP):
            __default_colormap = "jet"
        else:
            __default_colormap = config_read(COLORMAP)
    return __default_colormap


def set_default_colormap(colormap):
    global __default_colormap
    __default_colormap = colormap
    config_write(COLORMAP, colormap)


__current_workspace_path = None


def get_current_workspace_path():
    global __current_workspace_path
    return __current_workspace_path


def set_current_workspace_path(path):
    global __current_workspace_path
    __current_workspace_path = path


def get_skip_version():
    if not config_exists(SKIPVERSION):
        return 0
    return int(get_config().Read(SKIPVERSION))


def set_skip_version(ver):
    global __is_headless
    get_config().Write(SKIPVERSION, str(ver))
    if not __is_headless:
        get_config().Flush()


__show_sampling = None


def get_show_sampling():
    global __show_sampling
    if __show_sampling is not None:
        return __show_sampling
    if not config_exists(SHOW_SAMPLING):
        __show_sampling = False
        return False
    return get_config().ReadBool(SHOW_SAMPLING)


def set_show_sampling(value):
    global __show_sampling, __is_headless
    get_config().WriteBool(SHOW_SAMPLING, bool(value))
    __show_sampling = bool(value)
    if not __is_headless:
        get_config().Flush()


__recent_files: dict = {}


def get_recent_files(category=""):
    global __recent_files
    if __recent_files.get(category, None) is None:
        __recent_files[category] = []
        for i in range(RECENT_FILE_COUNT):
            key = recent_file(i, category)
            try:
                if config_exists(key):
                    __recent_files[category].append(config_read(key))
            except:
                pass
    return __recent_files[category]


def add_recent_file(filename, category=""):
    recent_files = get_recent_files(category)
    filename = os.path.abspath(filename)
    if filename in recent_files:
        recent_files.remove(filename)
    recent_files.insert(0, filename)
    if len(recent_files) > RECENT_FILE_COUNT:
        del recent_files[-1]
    for i, filename in enumerate(recent_files):
        config_write(recent_file(i, category), filename)


__plugin_directory = None
__conserve_memory = None


def get_plugin_directory():
    global __plugin_directory

    if __plugin_directory is not None:
        return __plugin_directory

    if config_exists(PLUGIN_DIRECTORY):
        __plugin_directory = config_read(PLUGIN_DIRECTORY)
    elif get_headless():
        return None
    elif config_exists("PluginDirectory"):
        # Import and store directory from CP3
        __plugin_directory = config_read("PluginDirectory")
        set_plugin_directory(__plugin_directory)
    else:
        import wx

        if wx.GetApp() is not None:
            __plugin_directory = os.path.join(
                wx.StandardPaths.Get().GetUserDataDir(), "plugins"
            )
    return __plugin_directory


def set_plugin_directory(value, globally=True):
    global __plugin_directory

    __plugin_directory = value
    if globally:
        config_write(PLUGIN_DIRECTORY, value)


__data_file = None


def get_data_file():
    """Get the path to the LoadData data file specified on the command-line"""
    global __data_file
    return __data_file


def set_data_file(path):
    global __data_file
    __data_file = path


def standardize_default_folder_names(setting_values, slot):
    if setting_values[slot] in list(FOLDER_CHOICE_TRANSLATIONS.keys()):
        replacement = FOLDER_CHOICE_TRANSLATIONS[setting_values[slot]]
    elif (
        setting_values[slot].startswith("Default Image")
        or setting_values[slot].startswith("Default image")
        or setting_values[slot].startswith("Default input")
    ):
        replacement = DEFAULT_INPUT_FOLDER_NAME
    elif setting_values[slot].startswith("Default output"):
        replacement = DEFAULT_OUTPUT_FOLDER_NAME
    else:
        replacement = setting_values[slot]
    setting_values = setting_values[:slot] + [replacement] + setting_values[slot + 1 :]
    return setting_values


__cpfigure_position = (-1, -1)


def get_next_cpfigure_position(update_next_position=True):
    global __cpfigure_position
    pos = __cpfigure_position
    if update_next_position:
        update_cpfigure_position()
    return pos


def reset_cpfigure_position():
    global __cpfigure_position
    __cpfigure_position = (-1, -1)


def update_cpfigure_position():
    """Called by get_next_cpfigure_position to update the screen position at
    which the next figure frame will be drawn.
    """
    global __cpfigure_position
    import wx

    win_size = (600, 400)
    try:
        disp = wx.GetDisplaySize()
    except:
        disp = (800, 600)
    if __cpfigure_position[0] + win_size[0] > disp[0]:
        __cpfigure_position = (-1, __cpfigure_position[1])
    if __cpfigure_position[1] + win_size[1] > disp[1]:
        __cpfigure_position = (-1, -1)
    else:
        # These offsets could be set in the preferences UI
        __cpfigure_position = (
            __cpfigure_position[0] + 120,
            __cpfigure_position[1] + 24,
        )


def get_telemetry():
    if not config_exists(TELEMETRY):
        return True

    return get_config().ReadBool(TELEMETRY)


def set_telemetry(val):
    global __is_headless
    get_config().WriteBool(TELEMETRY, val)
    if not __is_headless:
        get_config().Flush()


def get_telemetry_prompt():
    if not config_exists(TELEMETRY_PROMPT):
        return True

    return get_config().ReadBool(TELEMETRY_PROMPT)


def set_telemetry_prompt(val):
    global __is_headless
    get_config().WriteBool(TELEMETRY_PROMPT, val)
    if not __is_headless:
        get_config().Flush()


def get_check_update():
    if not config_exists(CHECK_FOR_UPDATES):
        return "Never"
    return get_config().Read(CHECK_FOR_UPDATES)


def get_check_update_bool():
    if not config_exists(CHECK_FOR_UPDATES):
        return True
    update_str = get_config().Read(CHECK_FOR_UPDATES)
    if update_str == "Disabled":
        return False
    else:
        return True


def set_check_update(val):
    if str(val) == "False":
        val = "Disabled"
    elif str(val) == "True":
        val = "Never"
    global __is_headless
    get_config().Write(CHECK_FOR_UPDATES, val)
    if not __is_headless:
        get_config().Flush()


def get_startup_blurb():
    if not config_exists(STARTUPBLURB):
        return True
    return get_config().ReadBool(STARTUPBLURB)


def set_startup_blurb(val):
    global __is_headless
    get_config().WriteBool(STARTUPBLURB, val)
    if not __is_headless:
        get_config().Flush()


def get_primary_outline_color():
    default = (0, 255, 0)
    if not config_exists(PRIMARY_OUTLINE_COLOR):
        return tuple_to_color(default)
    return tuple_to_color(config_read(PRIMARY_OUTLINE_COLOR).split(","))


def set_primary_outline_color(color):
    config_write(PRIMARY_OUTLINE_COLOR, ",".join([str(x) for x in color.Get()]))


def get_secondary_outline_color():
    default = (255, 0, 255)
    if not config_exists(SECONDARY_OUTLINE_COLOR):
        return tuple_to_color(default)
    return tuple_to_color(config_read(SECONDARY_OUTLINE_COLOR).split(","))


def set_secondary_outline_color(color):
    config_write(SECONDARY_OUTLINE_COLOR, ",".join([str(x) for x in color.Get()]))


def get_tertiary_outline_color():
    default = (255, 128, 0)
    if not config_exists(TERTIARY_OUTLINE_COLOR):
        return tuple_to_color(default)
    return tuple_to_color(config_read(TERTIARY_OUTLINE_COLOR).split(","))


def set_tertiary_outline_color(color):
    config_write(TERTIARY_OUTLINE_COLOR, ",".join([str(x) for x in color.Get()]))


__has_reported_jvm_error = False


def get_report_jvm_error():
    """Return true if user still wants to report a JVM error"""
    if __has_reported_jvm_error:
        return False
    if not config_exists(JVM_ERROR):
        return True
    return config_read(JVM_ERROR) == "True"


def set_report_jvm_error(should_report):
    config_write(JVM_ERROR, "True" if should_report else "False")


def set_has_reported_jvm_error():
    """Call this to remember that we showed the user the JVM error"""
    global __has_reported_jvm_error
    __has_reported_jvm_error = True


# "Analysis complete" preference
__show_analysis_complete_dlg = None


def get_show_analysis_complete_dlg():
    """Return true if the user wants to see the "analysis complete" dialog"""
    global __show_analysis_complete_dlg
    if __show_analysis_complete_dlg is not None:
        return __show_analysis_complete_dlg
    if not config_exists(SHOW_ANALYSIS_COMPLETE_DLG):
        return True
    return config_read(SHOW_ANALYSIS_COMPLETE_DLG) == "True"


def set_show_analysis_complete_dlg(value):
    """Set the "show analysis complete" flag"""
    global __show_analysis_complete_dlg
    __show_analysis_complete_dlg = value
    config_write(SHOW_ANALYSIS_COMPLETE_DLG, "True" if value else "False")


# "Existing test mode" preference
__show_exiting_test_mode_dlg = None


def get_show_exiting_test_mode_dlg():
    """Return true if the user wants to see the "exiting test mode" dialog"""
    global __show_exiting_test_mode_dlg
    if __show_exiting_test_mode_dlg is not None:
        return __show_exiting_test_mode_dlg
    if not config_exists(SHOW_EXITING_TEST_MODE_DLG):
        return True
    return config_read(SHOW_EXITING_TEST_MODE_DLG) == "True"


def set_show_exiting_test_mode_dlg(value):
    """Set the "exiting test mode" flag"""
    global __show_exiting_test_mode_dlg
    __show_exiting_test_mode_dlg = value
    config_write(SHOW_EXITING_TEST_MODE_DLG, "True" if value else "False")


# "Report bad sizes" preference
__show_report_bad_sizes_dlg = None


def get_show_report_bad_sizes_dlg():
    """Return true if the user wants to see the "report bad sizes" dialog"""
    global __show_report_bad_sizes_dlg
    if __show_report_bad_sizes_dlg is not None:
        return __show_report_bad_sizes_dlg
    if not config_exists(SHOW_BAD_SIZES_DLG):
        return True
    return config_read(SHOW_BAD_SIZES_DLG) == "True"


def set_show_report_bad_sizes_dlg(value):
    """Set the "exiting test mode" flag"""
    global __show_report_bad_sizes_dlg
    __show_report_bad_sizes_dlg = value
    config_write(SHOW_BAD_SIZES_DLG, "True" if value else "False")


__workspace_file = None


def get_workspace_file():
    """Return the path to the workspace file"""
    global __workspace_file
    if __workspace_file is not None:
        return __workspace_file
    if not config_exists(WORKSPACE_FILE):
        return None
    __workspace_file = config_read(WORKSPACE_FILE)
    return __workspace_file


def set_workspace_file(path, permanently=True):
    """Set the path to the workspace file

    path - path to the file

    permanently - True to write it to the configuration, False if the file
                  should only be set for the running instance (e.g., as a
                  command-line parameter for a scripted run)
    """
    global __workspace_file
    __workspace_file = path
    if permanently:
        add_recent_file(path, WORKSPACE_FILE)
        config_write(WORKSPACE_FILE, path)


###########################################
#
# OMERO logon credentials
#
###########################################

__omero_server = None
__omero_port = None
__omero_user = None
__omero_session_id = None


def get_omero_server():
    """Get the DNS name of the Omero server"""
    global __omero_server
    if __omero_server is None:
        if not config_exists(OMERO_SERVER):
            return None
        __omero_server = config_read(OMERO_SERVER)
    return __omero_server


def set_omero_server(omero_server, globally=True):
    """Set the DNS name of the Omero server"""
    global __omero_server
    __omero_server = omero_server
    if globally:
        config_write(OMERO_SERVER, omero_server)


def get_omero_port():
    """Get the port used to connect to the Omero server"""
    global __omero_port
    if __omero_port is None:
        if not config_exists(OMERO_PORT):
            return 4064
        try:
            __omero_port = int(config_read(OMERO_PORT))
        except:
            return 4064
    return __omero_port


def set_omero_port(omero_port, globally=True):
    """Set the port used to connect to the Omero server"""
    global __omero_port
    __omero_port = omero_port
    if globally:
        config_write(OMERO_PORT, str(omero_port))


def get_omero_user():
    """Get the Omero user name"""
    global __omero_user
    if __omero_user is None:
        if not config_exists(OMERO_USER):
            return None
        __omero_user = config_read(OMERO_USER)
    return __omero_user


def set_omero_user(omero_user, globally=True):
    """Set the Omero user name"""
    global __omero_user
    __omero_user = omero_user
    if globally:
        config_write(OMERO_USER, omero_user)


def get_omero_session_id():
    """Get the session ID to use to communicate to Omero"""
    global __omero_session_id
    if __omero_session_id is None:
        if not config_exists(OMERO_SESSION_ID):
            return None
        __omero_session_id = config_read(OMERO_SESSION_ID)
    return __omero_session_id


def set_omero_session_id(omero_session_id, globally=True):
    """Set the Omero session ID"""
    global __omero_session_id
    __omero_session_id = omero_session_id
    if globally:
        config_write(OMERO_SESSION_ID, omero_session_id)


def default_max_workers():
    return psutil.cpu_count(logical=False)


__max_workers = None


def get_max_workers():
    """Get the maximum number of worker processes allowed during analysis"""
    global __max_workers
    if __max_workers is not None:
        return __max_workers
    default = default_max_workers()
    if config_exists(MAX_WORKERS):
        __max_workers = get_config().ReadInt(MAX_WORKERS, default)
        return __max_workers
    return default


def set_max_workers(value):
    """Set the maximum number of worker processes allowed during analysis"""
    global __max_workers, __is_headless
    get_config().WriteInt(MAX_WORKERS, value)
    __max_workers = value
    if not __is_headless:
        get_config().Flush()


__temp_dir = None


def get_temporary_directory():
    """Get the directory to be used for temporary files

    The default is whatever is returned by tempfile.gettempdir()
    (see http://docs.python.org/2/library/tempfile.html#tempfile.gettempdir)
    """
    global __temp_dir
    if __temp_dir is not None:
        pass
    elif config_exists(TEMP_DIR):
        path = config_read(TEMP_DIR)
        if os.path.isdir(path):
            __temp_dir = path
            return __temp_dir
    __temp_dir = tempfile.gettempdir()
    return __temp_dir


def set_temporary_directory(tempdir, globally=False):
    """Set the directory to be used for temporary files

    tempdir - pathname of the directory
    """
    global __temp_dir
    if globally:
        config_write(TEMP_DIR, tempdir)
    __temp_dir = tempdir
    tempfile.tempdir = tempdir


__progress_data = threading.local()
__progress_data.last_report = time.time()
__progress_data.callbacks = None
__interpolation_mode = None


def get_interpolation_mode():
    """Get the interpolation mode for matplotlib

    Returns one of IM_NEAREST, IM_BILINEAR or IM_BICUBIC
    """
    global __interpolation_mode
    if __interpolation_mode is not None:
        return __interpolation_mode
    if config_exists(INTERPOLATION_MODE):
        __interpolation_mode = config_read(INTERPOLATION_MODE)
    else:
        __interpolation_mode = IM_NEAREST
    return __interpolation_mode


def set_interpolation_mode(value):
    global __interpolation_mode
    __interpolation_mode = value
    config_write(INTERPOLATION_MODE, value)


__intensity_mode = None


def get_intensity_mode():
    """Get the intensity scaling mode for matplotlib

    Returns one of INTENSITY_MODE_RAW, INTENSITY_MODE_NORMAL, INTENSITY_MODE_LOG
    """
    global __intensity_mode
    if __intensity_mode is not None:
        return __intensity_mode
    if config_exists(INTENSITY_MODE):
        __intensity_mode = config_read(INTENSITY_MODE)
    else:
        __intensity_mode = INTENSITY_MODE_NORMAL
    return __intensity_mode


def set_intensity_mode(value):
    """Set the intensity scaling mode for matplotlib"""
    global __intensity_mode
    __intensity_mode = value
    config_write(INTENSITY_MODE, value)


__save_pipeline_with_project = None


def get_save_pipeline_with_project():
    global __save_pipeline_with_project
    if __save_pipeline_with_project is None:
        if config_exists(SAVE_PIPELINE_WITH_PROJECT):
            __save_pipeline_with_project = config_read(SAVE_PIPELINE_WITH_PROJECT)
        else:
            __save_pipeline_with_project = SPP_NEITHER
    return __save_pipeline_with_project


def set_save_pipeline_with_project(value):
    global __save_pipeline_with_project
    __save_pipeline_with_project = value
    config_write(SAVE_PIPELINE_WITH_PROJECT, value)


__allow_schema_write = True


def get_allow_schema_write():
    """Returns True if ExportToDatabase is allowed to write the MySQL schema

    For cluster operation without CreateBatchFiles, it's inappropriate to
    have multiple processes overwrite the database schema. Although
    CreateBatchFiles is suggested for this scenario, we put this switch in
    to support disabling schema writes from the command line.
    """
    return __allow_schema_write


def set_allow_schema_write(value):
    """Allow or disallow database schema writes

    value - True to allow writes (the default) or False to prevent
            ExportToDatabase from writing the schema.

    For cluster operation without CreateBatchFiles, it's inappropriate to
    have multiple processes overwrite the database schema. Although
    CreateBatchFiles is suggested for this scenario, we put this switch in
    to support disabling schema writes from the command line.
    """
    global __allow_schema_write
    __allow_schema_write = value


__filename_re_guess_file = None


def get_filename_re_guess_file():
    """The path to the file that contains filename regular expression guesses

    The file given by this preference is an optional file that contains
    possible regular expression patterns to match against file names.
    """
    global __filename_re_guess_file
    if __filename_re_guess_file is None:
        if config_exists(FILENAME_RE_GUESSES_FILE):
            __filename_re_guess_file = config_read(FILENAME_RE_GUESSES_FILE)
    return __filename_re_guess_file


def set_filename_re_guess_file(value):
    """Set the path to the filename regular expression guess file"""
    global __filename_re_guess_file
    __filename_re_guess_file = value
    config_write(FILENAME_RE_GUESSES_FILE, value)


__pathname_re_guess_file = None


def get_pathname_re_guess_file():
    """The path to the file that contains pathname regular expression guesses

    The file given by this preference is an optional file that contains
    possible regular expression patterns to match against path names.
    """
    global __pathname_re_guess_file
    if __pathname_re_guess_file is None:
        if config_exists(PATHNAME_RE_GUESSES_FILE):
            __pathname_re_guess_file = config_read(PATHNAME_RE_GUESSES_FILE)
    return __pathname_re_guess_file


def set_pathname_re_guess_file(value):
    """Set the path to the pathname regular expression guess file"""
    global __pathname_re_guess_file
    __pathname_re_guess_file = value
    config_write(PATHNAME_RE_GUESSES_FILE, value)


__image_set_filename = None

__wants_pony = None


def get_wants_pony():
    """

    :return:

    """
    global __wants_pony

    if __wants_pony is not None:
        return __wants_pony
    elif config_exists("Pony"):
        return config_read("Pony").lower() == "yes"
    else:
        return False


def set_wants_pony(wants_pony):
    """

    :param wants_pony:

    """
    global __wants_pony

    __wants_pony = wants_pony

    config_write("Pony", "yes" if wants_pony else "no")


def set_image_set_file(filename):
    """Record the name of the image set that should be loaded upon startup"""
    global __image_set_filename
    __image_set_filename = filename


def clear_image_set_file():
    """Remove the recorded image set file name

    Call this after loading the image set file to cancel reloading of the
    file during subsequent operations.
    """
    global __image_set_filename
    __image_set_filename = None


def get_image_set_file():
    """Recover the name of the image set file to use to populate the file list

    Returns either None or the name of the file to use. For the UI, the
    file list should be loaded and clear_image_set_file() should be called,
    for headless, the file list should be loaded after the pipeline has been
    loaded.
    """
    return __image_set_filename


__choose_image_set_frame_size = None


def get_choose_image_set_frame_size():
    """Return the size (w, h) for the "Choose image set" dialog frame"""
    global __choose_image_set_frame_size
    if __choose_image_set_frame_size is None:
        if config_exists(CHOOSE_IMAGE_SET_FRAME_SIZE):
            s = config_read(CHOOSE_IMAGE_SET_FRAME_SIZE)
            __choose_image_set_frame_size = tuple(
                [int(_.strip()) for _ in s.split(",", 1)]
            )
    return __choose_image_set_frame_size


def set_choose_image_set_frame_size(w, h):
    """Set the size of the "Choose image set" dialog frame"""
    global __choose_image_set_frame_size
    __choose_image_set_frame_size = (w, h)
    config_write(CHOOSE_IMAGE_SET_FRAME_SIZE, "%d,%d" % (w, h))


__normalization_factor = None


def get_normalization_factor():
    global __normalization_factor
    if __normalization_factor is not None:
        return __normalization_factor
    if config_exists(NORMALIZATION_FACTOR):
        __normalization_factor = config_read(NORMALIZATION_FACTOR)
    else:
        __normalization_factor = "1.0"
    return __normalization_factor


def set_normalization_factor(normalization_factor):
    global __normalization_factor
    try:
        float(normalization_factor)
    except ValueError:
        print(
            f"Unable to set {NORMALIZATION_FACTOR} to {normalization_factor}, value must be a number."
        )
        return
    __normalization_factor = normalization_factor
    config_write(NORMALIZATION_FACTOR, normalization_factor)


def get_conserve_memory():
    global __conserve_memory
    if __conserve_memory is not None:
        return __conserve_memory in (True, "True")
    if not config_exists(CONSERVE_MEMORY):
        return False
    return get_config().ReadBool(CONSERVE_MEMORY)


def set_conserve_memory(val, globally=True):
    global __conserve_memory
    __conserve_memory = val
    if globally:
        config_write(CONSERVE_MEMORY, val)


def add_progress_callback(callback):
    """Add a callback function that listens to progress calls

    The progress indicator is designed to monitor progress of operations
    on the user interface thread. The model is that operations are nested
    so that both an operation and sub-operation can report their progress.
    An operation reports its initial progress and is pushed onto the
    stack at that point. When it reports 100% progress, it's popped from
    the stack.

    callback - callback function with signature of
               fn(operation_id, progress, message)
               where operation_id names the instance of the operation being
               performed (e.g., a UUID), progress is a number between 0 and 1
               where 1 indicates that the operation has completed and
               message is the message to show.

               Call the callback with operation_id = None to pop the operation
               stack after an exception.

    Note that the callback must remain in-scope. For example:

    class Foo():
       def callback(operation_id, progress, message):
          ...

    works but

    class Bar():
        def __init__(self):
            def callback(operation_id, progress, message):
                ...

    does not work because the reference is lost when __init__ returns.
    """
    global __progress_data
    if __progress_data.callbacks is None:
        __progress_data.callbacks = weakref.WeakSet()
    __progress_data.callbacks.add(callback)


def remove_progress_callback(callback):
    global __progress_data
    if __progress_data.callbacks is not None and callback in __progress_data.callbacks:
        __progress_data.callbacks.remove(callback)


def report_progress(operation_id, progress, message):
    """Report progress to all callbacks registered on the caller's thread

    operation_id - ID of operation being performed

    progress - a number between 0 and 1 indicating the extent of progress.
               None indicates indeterminate operation duration. 0 should be
               reported at the outset and 1 at the end.

    message - an informative message.
    """
    global __progress_data
    if __progress_data.callbacks is None:
        return
    t = time.time()
    if progress in (None, 0, 1) or t - __progress_data.last_report > 1:
        for callback in __progress_data.callbacks:
            callback(operation_id, progress, message)
        __progress_data.last_report = time.time()


def map_report_progress(fn_map, fn_report, sequence, freq=None):
    """Apply a mapping function to a sequence, reporting progress

    fn_map - function that maps members of the sequence to members of the output

    fn_report - function that takes a sequence member and generates an
                informative string

    freq - report on mapping every N items. Default is to report 100 or less
           times.
    """
    n_items = len(sequence)
    if n_items == 0:
        return []
    if freq is None:
        if n_items < 100:
            freq = 1
        else:
            freq = (n_items + 99) / 100
    output = []
    uid = uuid.uuid4()
    for i in range(0, n_items, freq):
        report_progress(uuid, float(i) / n_items, fn_report(sequence[i]))
        output += list(map(fn_map, sequence[i : i + freq]))
    report_progress(uuid, 1, "Done")
    return output


def cancel_progress():
    """Cancel all progress indicators

    for instance, after an exception is thrown that bubbles to the top.
    """
    report_progress(None, None, None)

__always_continue = None
def get_always_continue():
    global __always_continue
    if __always_continue is not None:
        return __always_continue in (True, "True")
    if not config_exists(ALWAYS_CONTINUE):
        return False
    return get_config().ReadBool(ALWAYS_CONTINUE)


def set_always_continue(val, globally=True):
    global __always_continue
    __always_continue = val
    if globally:
        config_write(ALWAYS_CONTINUE, val)

__widget_inspector = None
# global_only - only return True if local is set
# ie ignore config settings
def get_widget_inspector(global_only=False):
    global __widget_inspector
    if not global_only and __widget_inspector is not None:
        return __widget_inspector == True
    if not config_exists(WIDGET_INSPECTOR):
        return False
    return get_config().ReadBool(WIDGET_INSPECTOR)


def set_widget_inspector(val, globally=True):
    global __widget_inspector
    __widget_inspector = val
    if globally:
        config_write(WIDGET_INSPECTOR, val)



from ._do_something import DoSomething


class ImageSetDisplay(DoSomething):
    """A button that refreshes the image set display when pressed

    """

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(
            args[0], args[1], None, *args[:2], **kwargs
        )



import gc
import threading
import uuid

from ._runner import Runner
from ..measurement import Measurements


class Analysis:
    """An Analysis is the application of a particular pipeline of modules to a
    set of images to produce measurements.

    Multiprocessing for analyses is handled by multiple layers of threads and
    processes, to keep the GUI responsive and simplify the code.  Threads and
    processes are organized as below.  Display/Interaction requests and
    Exceptions are sent directly to the pipeline listener.

    +------------------------------------------------+
    |           CellProfiler GUI/WX thread           |
    |                                                |
    +- Analysis() methods down,  Events/Requests up -+
    |                                                |
    |       AnalysisRunner.interface() thread        |
    |                                                |
    +----------------  Queues  ----------------------+
    |                                                |
    |  AnalysisRunner.jobserver()/announce() threads |
    |                                                |
    +----------------------------------------------- +
    |              zmqrequest.Boundary()             |
    +---------------+----------------+---------------+
    |     Worker    |     Worker     |   Worker      |
    +---------------+----------------+---------------+

    Workers are managed by class variables in the AnalysisRunner.
    """

    def __init__(self, pipeline, initial_measurements=None):
        """create an Analysis applying pipeline to a set of images, writing out
        to measurements_filename, optionally starting with previous
        measurements."""
        self.pipeline = pipeline
        initial_measurements = Measurements(copy=initial_measurements)
        self.initial_measurements_buf = initial_measurements.file_contents()
        initial_measurements.close()
        self.debug_mode = False
        self.analysis_in_progress = False
        self.runner = None

        self.runner_lock = threading.Lock()  # defensive coding purposes

    def start(self, analysis_event_callback, num_workers=None, overwrite=True):
        """Start the analysis runner

        analysis_event_callback - callback from runner to UI thread for
                                  event progress and UI handlers

        num_workers - # of worker processes to instantiate, default is # of cores

        overwrite - True (default) to process all image sets, False to only
                    process incomplete ones (or incomplete groups if grouping)
        """
        with self.runner_lock:
            assert not self.analysis_in_progress
            self.analysis_in_progress = uuid.uuid1().hex

            self.runner = Runner(
                self.analysis_in_progress,
                self.pipeline,
                self.initial_measurements_buf,
                analysis_event_callback,
            )
            self.runner.start(num_workers=num_workers, overwrite=overwrite)
            return self.analysis_in_progress

    def pause(self):
        with self.runner_lock:
            assert self.analysis_in_progress
            self.runner.pause()

    def resume(self):
        with self.runner_lock:
            assert self.analysis_in_progress
            self.runner.resume()

    def cancel(self):
        with self.runner_lock:
            if not self.analysis_in_progress:
                return
            self.analysis_in_progress = False
            self.runner.cancel()
            self.runner = None
            """
            Explicitly call GC here so that the old runner's threads and 
            queues get deleted before any new analysis can be started.
            On rare occasions CP can hang when threads get GC'd during a run
            """
            gc.collect()

    def check_running(self):
        """Verify that an analysis is running, allowing the GUI to recover even
        if the AnalysisRunner fails in some way.

        Returns True if analysis is still running (threads are still alive).
        """
        with self.runner_lock:
            if self.analysis_in_progress:
                return self.runner.check()
            return False



"""_workspace.py - the workspace for an imageset
"""

import io
import logging
import os

import h5py

import cellprofiler_core.utilities.measurement
from ._disposition_changed_event import DispositionChangedEvent
from ..constants.measurement import EXPERIMENT
from ..constants.workspace import DISPOSITION_CONTINUE
from ..utilities.hdf5_dict import HDF5FileList


LOGGER = logging.getLogger(__name__)

class Workspace:
    """The workspace contains the processing information and state for
    a pipeline run on an image set
    """

    def __init__(
        self,
        pipeline,
        module,
        image_set,
        object_set,
        measurements,
        image_set_list,
        frame=None,
        create_new_window=False,
        outlines=None,
    ):
        """Workspace constructor

        pipeline          - the pipeline of modules being run
        module            - the current module to run (a CPModule instance)
        image_set         - the set of images available for this iteration
                            (a cpimage.ImageSet instance)
        object_set        - an object.ObjectSet instance
        image_set_list    - the list of all images
        frame             - the application's frame, or None for no display
        create_new_window - True to create another frame, even if one is open
                            False to reuse the current frame.
        """
        if outlines is None:
            outlines = {}
        self.__pipeline = pipeline
        self.__module = module
        self.__image_set = image_set
        self.__object_set = object_set
        self.__measurements = measurements
        self.__image_set_list = image_set_list
        self.__frame = frame
        self.__do_show = frame is not None
        self.__outlines = outlines
        self.__windows_used = []
        self.__create_new_window = create_new_window
        self.__grid = {}
        self.__disposition = DISPOSITION_CONTINUE
        self.__disposition_listeners = []
        self.__in_background = (
            False  # controls checks for calls to create_or_find_figure()
        )
        self.__filename = None
        self.__file_list = None
        self.__loading = False
        if measurements is not None:
            self.set_file_list(HDF5FileList(measurements.hdf5_dict.hdf5_file))
        self.__notification_callbacks = []

        self.interaction_handler = None
        self.post_run_display_handler = None
        self.post_group_display_handler = None
        self.cancel_handler = None

        class DisplayData(object):
            pass

        self.display_data = DisplayData()
        """Object into which the module's run() method can stuff items
        that must be available later for display()."""

    def refresh(self):
        """Refresh any windows created during use"""
        for window in self.__windows_used:
            window.figure.canvas.draw()

    def get_windows_used(self):
        return self.__windows_used

    def get_pipeline(self):
        """Get the pipeline being run"""
        return self.__pipeline

    pipeline = property(get_pipeline)

    def get_image_set(self):
        """The image set is the set of images currently being processed
        """
        return self.__image_set

    def set_image_set_for_testing_only(self, image_set_number):
        self.__image_set = self.image_set_list.get_image_set(image_set_number)

    image_set = property(get_image_set)

    def get_image_set_list(self):
        """The list of all image sets"""
        return self.__image_set_list

    image_set_list = property(get_image_set_list)

    def get_object_set(self):
        """The object set is the set of image labels for the current image set
        """
        return self.__object_set

    object_set = property(get_object_set)

    def get_objects(self, objects_name):
        """Return the objects.Objects instance for the given name.

        objects_name - the name of the objects to retrieve
        """
        return self.object_set.get_objects(objects_name)

    def get_measurements(self):
        """The measurements contain measurements made on images and objects
        """
        return self.__measurements

    measurements = property(get_measurements)

    def add_measurement(self, object_name, feature_name, data):
        """Add a measurement to the workspace's measurements

        object_name - name of the objects measured or 'Image'
        feature_name - name of the feature measured
        data - the result of the measurement
        """
        self.measurements.add_measurement(object_name, feature_name, data)

    def get_file_list(self):
        """The user-curated list of files"""
        return self.__file_list

    def set_file_list(self, file_list):
        """Set the file list

        A caller can set the file list to the file list in some other
        workspace. This lets a single, sometimes very bulky file list be
        used without copying it to a measurements file.
        """
        if self.__file_list is not None:
            self.__file_list.remove_notification_callback(self.__on_file_list_changed)
        self.__file_list = file_list
        if file_list is not None:
            self.__file_list.add_notification_callback(self.__on_file_list_changed)

    file_list = property(get_file_list)

    def get_grid(self, grid_name):
        """Return a grid with the given name"""
        if grid_name not in self.__grid:
            raise ValueError("Could not find grid %s" % grid_name)
        return self.__grid[grid_name]

    def set_grids(self, last=None):
        """Initialize the grids for an image set

        last - none if first in image set or the return value from
               this method.
        returns a grid dictionary
        """
        if last is None:
            last = {}
        self.__grid = last
        return self.__grid

    def set_grid(self, grid_name, grid_info):
        """Add a grid to the workspace"""
        self.__grid[grid_name] = grid_info

    def get_frame(self):
        """The frame is CellProfiler's gui window

        If the frame is present, a module should do its display
        """
        if self.__do_show:
            return self.__frame
        return None

    frame = property(get_frame)

    def show_frame(self, do_show):
        self.__do_show = do_show

    def get_display(self):
        """True to provide a gui display"""
        return self.__frame is not None

    display = property(get_display)

    def get_in_background(self):
        return self.__in_background

    def set_in_background(self, val):
        self.__in_background = val

    in_background = property(get_in_background, set_in_background)

    def get_outline_names(self):
        """The names of outlines of objects"""
        return list(self.__outlines.keys())

    def add_outline(self, name, outline):
        """Add an object outline to the workspace"""
        self.__outlines[name] = outline

    def get_outline(self, name):
        """Get a named outline"""
        return self.__outlines[name]

    def get_module(self):
        """Get the module currently being run"""
        return self.__module

    module = property(get_module)

    def set_module(self, module):
        """Set the module currently being run"""
        self.__module = module

    def interaction_request(self, module, *args, **kwargs):
        """make a request for GUI interaction via a pipeline event

        module - target module for interaction request

        headless_ok - True if the interaction request can be made in
                      a headless context. An example is synchronized access to
                      a shared resource which must be coordinated among all
                      workers.
        """
        # See also:
        # main().interaction_handler() in __init__.py
        # PipelineController.module_interaction_request() in pipelinecontroller.py
        import cellprofiler_core.preferences

        if "headless_ok" in kwargs:
            tmp = kwargs.copy()
            del tmp["headless_ok"]
            headless_ok = kwargs["headless_ok"]
            kwargs = tmp
        else:
            headless_ok = False
        if self.interaction_handler is None:
            if cellprofiler_core.preferences.get_headless() and not headless_ok:
                raise self.NoInteractionException()
            else:
                return module.handle_interaction(*args, **kwargs)
        else:
            return self.interaction_handler(module, *args, **kwargs)

    def cancel_request(self):
        """Make a request to cancel an ongoing analysis"""
        if self.cancel_handler is None:
            raise self.NoInteractionException()
        self.cancel_handler()

    def post_group_display(self, module):
        """Perform whatever post-group module display is necessary

        module - module being run
        """
        if self.post_group_display_handler is not None:
            self.post_group_display_handler(
                module, self.display_data, self.measurements.image_set_number
            )
        elif self.frame is not None:
            figure = self.get_module_figure(
                module, self.measurements.image_set_number, self.frame
            )
            module.display_post_group(self, figure)

    def post_run_display(self, module):
        """Perform whatever post-run module display is necessary

        module - module being run
        """
        if self.post_run_display_handler is not None:
            self.post_run_display_handler(self, module)
        elif self.frame is not None:
            figure = self.get_module_figure(
                module, self.measurements.image_set_count + 1, self.frame
            )
            module.display_post_run(self, figure)

    @property
    def is_last_image_set(self):
        return self.measurements.image_set_number == self.image_set_list.count() - 1

    def get_disposition(self):
        """How to proceed with the pipeline

        One of the following values:
        DISPOSITION_CONTINUE - continue to execute the pipeline
        DISPOSITION_PAUSE - wait until the status changes before executing
                            the next module
        DISPOSITION_CANCEL - stop running the pipeline
        DISPOSITION_SKIP - skip the rest of this image set
        """
        return self.__disposition

    def set_disposition(self, disposition):
        self.__disposition = disposition
        event = DispositionChangedEvent(disposition)
        for listener in self.__disposition_listeners:
            listener(event)

    disposition = property(get_disposition, set_disposition)

    def add_disposition_listener(self, listener):
        self.__disposition_listeners.append(listener)

    class NoInteractionException(Exception):
        pass

    def load(self, filename, load_pipeline):
        """Load a workspace from a .cpi file

        filename - path to file to load

        load_pipeline - true to load the pipeline from the file, false to
                        use the current pipeline.
        """
        import shutil
        from ..constants.pipeline import M_DEFAULT_OUTPUT_FOLDER
        from ..constants.pipeline import M_DEFAULT_INPUT_FOLDER
        from ..constants.pipeline import M_PIPELINE
        import cellprofiler_core.measurement
        from cellprofiler_core.preferences import (
            set_default_image_directory,
            set_default_output_directory,
        )

        image_set_and_measurements_are_same = False
        if self.__measurements is not None:
            image_set_and_measurements_are_same = id(self.__measurements) == id(
                self.__image_set
            )
            self.close()
        self.__loading = True
        try:
            #
            # Copy the file to a temporary location before opening
            #
            (
                fd,
                self.__filename,
            ) = cellprofiler_core.utilities.measurement.make_temporary_file()
            os.close(fd)

            shutil.copyfile(filename, self.__filename)

            self.__measurements = cellprofiler_core.measurement.Measurements(
                filename=self.__filename, mode="r+"
            )
            if self.__file_list is not None:
                self.__file_list.remove_notification_callback(
                    self.__on_file_list_changed
                )
            self.__file_list = HDF5FileList(self.measurements.hdf5_dict.hdf5_file)
            self.__file_list.add_notification_callback(self.__on_file_list_changed)
            if load_pipeline and self.__measurements.has_feature(
                EXPERIMENT, M_PIPELINE
            ):
                pipeline_txt = self.__measurements.get_experiment_measurement(
                    M_PIPELINE
                )
                # CP 3.1.8 cpproj (and possibly before) saved info in bytes; converting to python 3 string
                if type(pipeline_txt) == bytes:
                    pipeline_txt = (
                        pipeline_txt.decode("unicode_escape")
                        .encode("utf-8")
                        .replace(b"\\x00", b"")
                        .decode("unicode_escape")
                        .replace("ÿþ", "")
                    )
                if "\\n" in pipeline_txt:
                    # Loaded pipeline text from a pre-h5py 3 hdf5 will have escaped characters in need of decoding.
                    try:
                        pipeline_txt = pipeline_txt.encode(
                            "latin-1", "backslashreplace"
                        ).decode("unicode-escape")
                    except Exception as e:
                        print(
                            f"Unable to fully decode pipeline, you may encounter some errors. Issue was: {e}"
                        )
                self.pipeline.load(io.StringIO(pipeline_txt))
            elif load_pipeline:
                self.pipeline.clear()
            else:
                fd = io.StringIO()
                self.pipeline.dump(fd, save_image_plane_details=False)
                self.__measurements.add_experiment_measurement(
                    M_PIPELINE, fd.getvalue()
                )

            for feature, function in (
                (M_DEFAULT_INPUT_FOLDER, set_default_image_directory),
                (M_DEFAULT_OUTPUT_FOLDER, set_default_output_directory),
            ):
                if self.measurements.has_feature(EXPERIMENT, feature):
                    path = self.measurements[EXPERIMENT, feature]
                    if os.path.isdir(path):
                        function(path)
            if image_set_and_measurements_are_same:
                self.__image_set = self.__measurements

        finally:
            self.__loading = False
        self.notify(self.WorkspaceLoadedEvent(self))

    def create(self):
        """Create a new workspace file

        filename - name of the workspace file
        """
        from ..utilities.measurement import make_temporary_file
        from cellprofiler_core.measurement import Measurements

        if isinstance(self.measurements, Measurements):
            self.close()

        fd, self.__filename = make_temporary_file()
        self.__measurements = Measurements(filename=self.__filename, mode="w")
        os.close(fd)
        if self.__file_list is not None:
            self.__file_list.remove_notification_callback(self.__on_file_list_changed)
        self.__file_list = HDF5FileList(self.measurements.hdf5_dict.hdf5_file)
        self.__file_list.add_notification_callback(self.__on_file_list_changed)
        self.notify(self.WorkspaceCreatedEvent(self))

    def save(self, path):
        """Save the current workspace to the given path

        path - path to file to save

        Note: "saving" means copying the temporary workspace file
        """
        self.save_default_folders_to_measurements()
        self.measurements.flush()
        #
        # Note: shutil.copy and similar don't seem to work under Windows.
        #       I suspect that there's some file mapping magic that's
        #       causing problems because I found some internet postings
        #       where people tried to copy database files and failed.
        #       If you're thinking, "He didn't close the file", I did.
        #       shutil.copy creates a truncated file if you use it.
        #
        hdf5src = self.measurements.hdf5_dict.hdf5_file
        hdf5dest = h5py.File(path, mode="w")
        for key in hdf5src:
            obj = hdf5src[key]
            if isinstance(obj, h5py.Dataset):
                hdf5dest[key] = obj[()]
            else:
                hdf5src.copy(hdf5src[key], hdf5dest, key)
        for key in hdf5src.attrs:
            hdf5dest.attrs[key] = hdf5src.attrs[key]
        hdf5dest.close()

    def close(self):
        """Close the workspace and delete the temporary measurements file"""
        if self.measurements is not None and self.__filename is not None:
            self.measurements.close()
            os.unlink(self.__filename)

    def save_pipeline_to_measurements(self):
        from ..constants.pipeline import M_PIPELINE

        fd = io.StringIO()
        self.pipeline.dump(fd, save_image_plane_details=False)
        self.measurements.add_experiment_measurement(M_PIPELINE, fd.getvalue())
        self.measurements.flush()

    def save_default_folders_to_measurements(self):
        from ..constants.pipeline import M_DEFAULT_INPUT_FOLDER
        from ..constants.pipeline import M_DEFAULT_OUTPUT_FOLDER
        from cellprofiler_core.preferences import get_default_image_directory
        from cellprofiler_core.preferences import get_default_output_directory

        self.measurements.add_experiment_measurement(
            M_DEFAULT_INPUT_FOLDER, get_default_image_directory()
        )
        self.measurements.add_experiment_measurement(
            M_DEFAULT_OUTPUT_FOLDER, get_default_output_directory()
        )

    def invalidate_image_set(self):
        if not self.__loading:
            self.measurements.clear()
            self.save_pipeline_to_measurements()

    def refresh_image_set(self, force=False):
        """Refresh the image set if not present

        force - force a rewrite, even if the image set is cached

        This method executes Pipeline.prepare_run in order on self in order
        to write the image set image measurements to our internal
        measurements. If image set measurements are present, then we
        assume that the cache reflects pipeline + file list unless "force"
        is true.
        """
        if len(self.measurements.get_image_numbers()) == 0 or force:
            self.measurements.clear()
            self.save_pipeline_to_measurements()
            modules = self.pipeline.modules()
            stop_module = None
            if len(modules) > 1 and not modules[-1].is_load_module():
                for module, next_module in zip(
                    self.pipeline.modules()[:-1], self.pipeline.modules()[1:]
                ):
                    if module.is_load_module():
                        stop_module = next_module

            # TODO: Get rid of image_set_list
            no_image_set_list = self.image_set_list is None
            if no_image_set_list:
                from cellprofiler_core.image import ImageSetList

                self.__image_set_list = ImageSetList()
            try:
                result = self.pipeline.prepare_run(self, stop_module)
                return result
            except:
                LOGGER.error("Failed during prepare_run", exc_info=1)
                return False
            finally:
                if no_image_set_list:
                    self.__image_set_list = None
        return True

    def add_notification_callback(self, callback):
        """Add a callback that will be called on a workspace event

        Workspace events are load events, and file list events.

        callback - a function to be called when an event occurs. The signature
        is: callback(event)
        """
        self.__notification_callbacks.append(callback)

    def remove_notification_callback(self, callback):
        self.__notification_callbacks.remove(callback)

    def notify(self, event):
        for callback in self.__notification_callbacks:
            try:
                callback(event)
            except:
                LOGGER.error("Notification callback threw an exception", exc_info=1)

    def __on_file_list_changed(self):
        self.notify(self.WorkspaceFileListNotification(self))

    class WorkspaceEvent(object):
        """The base for any event sent to a workspace callback via Workspace.notify

        """

        def __init__(self, workspace):
            self.workspace = workspace

    class WorkspaceLoadedEvent(WorkspaceEvent):
        """Indicates that a workspace has been loaded

        When a workspace loads, the file list changes.
        """

        def __init__(self, workspace):
            super(self.__class__, self).__init__(workspace)

    class WorkspaceCreatedEvent(WorkspaceEvent):
        """Indicates that a blank workspace has been created"""

        def __init__(self, workspace):
            super(self.__class__, self).__init__(workspace)

    class WorkspaceFileListNotification(WorkspaceEvent):
        """Indicates that the workspace's file list changed"""

        def __init__(self, workspace):
            super(self.__class__, self).__init__(workspace)



import json


class HeadlessConfiguration:
    """
    This class functions as a configuration set for headless runs.
    The default config class is wx-based, which means we have to replace it
    here with this psuedo-replacement that has the same interface
    """

    def __init__(self):
        self.__preferences = {}

    def Read(self, kwd):
        return self.__preferences[kwd]

    def ReadInt(self, kwd, default=0):
        return int(self.__preferences.get(kwd, default))

    def ReadBool(self, kwd, default=False):
        return bool(self.__preferences.get(kwd, default))

    def Write(self, kwd, value):
        self.__preferences[kwd] = value

    # wx implements these for their own version of the "Config" object
    # Because this class is a mock config object without wx, we need to
    # make its interface the same
    WriteInt = Write
    WriteBool = Write

    def Exists(self, kwd):
        return kwd in self.__preferences

    def GetEntryType(self, kwd):
        """Get the data type of the registry key.

        Returns wx.Config.Type_String = 1
        """
        return 1

    def load_json(self, path):
        with open(path) as fd:
            config_dict = json.load(fd)
        self.__preferences.update(config_dict)



from ._crop_image_name import CropImageName
from ._external_image_name import ExternalImageName
from ._file_image_name import FileImageName
from ._image_name import ImageName
from ._outline_image_name import OutlineImageName



from .._number import Number


class Integer(Number):
    """
    A setting that allows only integer input

    Initializer:
    text - explanatory text for setting
    value - default value
    minval - minimum allowed value defaults to no minimum
    maxval - maximum allowed value defaults to no maximum
    """

    def str_to_value(self, str_value):
        return int(str_value)

    def value_to_str(self, value):
        return "%d" % value



from ._image_name import ImageName


class ExternalImageName(ImageName):
    """
    A setting that provides an image name where the image is loaded externally. (eg: from Java)
    """

    def __init__(self, text, value="Do not use", *args, **kwargs):
        kwargs = kwargs.copy()

        if "provided_attributes" not in kwargs:
            kwargs["provided_attributes"] = {}

        kwargs["provided_attributes"]["external_image"] = True

        super(ExternalImageName, self).__init__(text, value, *args, **kwargs)



from cellprofiler_core.setting.subscriber import LabelSubscriber

from ._image_segmentation import ImageSegmentation
from ...constants.measurement import COLTYPE_INTEGER
from ...constants.measurement import C_CHILDREN
from ...constants.measurement import C_NUMBER
from ...constants.measurement import C_PARENT
from ...constants.measurement import FF_CHILDREN_COUNT
from ...constants.measurement import FF_COUNT
from ...constants.measurement import FF_PARENT
from ...constants.measurement import FTR_OBJECT_NUMBER
from ...object import Objects


class ObjectProcessing(ImageSegmentation):
    category = "Object Processing"

    def add_measurements(
        self, workspace, input_object_name=None, output_object_name=None
    ):
        if input_object_name is None:
            input_object_name = self.x_name.value

        if output_object_name is None:
            output_object_name = self.y_name.value

        super(ObjectProcessing, self).add_measurements(workspace, output_object_name)

        objects = workspace.object_set.get_objects(output_object_name)

        parent_objects = workspace.object_set.get_objects(input_object_name)

        children_per_parent, parents_of_children = parent_objects.relate_children(
            objects
        )

        workspace.measurements.add_measurement(
            input_object_name,
            FF_CHILDREN_COUNT % output_object_name,
            children_per_parent,
        )

        workspace.measurements.add_measurement(
            output_object_name, FF_PARENT % input_object_name, parents_of_children,
        )

    def create_settings(self):
        super(ObjectProcessing, self).create_settings()

        self.x_name = LabelSubscriber(
            "Select the input object", doc="Select the object you want to use."
        )

    def display(self, workspace, figure):
        layout = (2, 1)

        figure.set_subplots(
            dimensions=workspace.display_data.dimensions, subplots=layout
        )

        figure.subplot_imshow_labels(
            image=workspace.display_data.x_data, title=self.x_name.value, x=0, y=0
        )

        figure.subplot_imshow_labels(
            image=workspace.display_data.y_data,
            sharexy=figure.subplot(0, 0),
            title=self.y_name.value,
            x=1,
            y=0,
        )

    def get_categories(self, pipeline, object_name):
        if object_name == self.x_name.value:
            return [C_CHILDREN]

        categories = super(ObjectProcessing, self).get_categories(pipeline, object_name)

        if object_name == self.y_name.value:
            return categories + [C_PARENT]

        return categories

    def get_measurement_columns(self, pipeline, additional_objects=None):
        if additional_objects is None:
            additional_objects = []
        object_names = [(self.x_name.value, self.y_name.value)] + additional_objects

        columns = [
            super(ObjectProcessing, self).get_measurement_columns(
                pipeline, output_object_name
            )
            + [
                (
                    input_object_name,
                    FF_CHILDREN_COUNT % output_object_name,
                    COLTYPE_INTEGER,
                ),
                (output_object_name, FF_PARENT % input_object_name, COLTYPE_INTEGER,),
            ]
            for (input_object_name, output_object_name) in object_names
        ]

        return sum(columns, [])

    def get_measurements(self, pipeline, object_name, category):
        if object_name == self.x_name.value and category == C_CHILDREN:
            return ["%s_Count" % self.y_name.value]

        if object_name == self.y_name.value:
            if category == C_NUMBER:
                return [FTR_OBJECT_NUMBER]

            if category == C_PARENT:
                return [self.x_name.value]

        return super(ObjectProcessing, self).get_measurements(
            pipeline, object_name, category
        )

    def run(self, workspace):
        x_name = self.x_name.value

        y_name = self.y_name.value

        objects = workspace.object_set

        x = objects.get_objects(x_name)

        dimensions = x.dimensions

        x_data = x.segmented

        args = (setting.value for setting in self.settings()[2:])

        y_data = self.function(x_data, *args)

        y = Objects()

        y.segmented = y_data

        y.parent_image = x.parent_image

        objects.add_objects(y, y_name)

        self.add_measurements(workspace)

        if self.show_window:
            workspace.display_data.x_data = x_data

            workspace.display_data.y_data = y_data

            workspace.display_data.dimensions = dimensions

    def volumetric(self):
        return True



from ._pathname import Pathname


class URL(Pathname):
    """
    A setting that displays a path name or URL
    """

    def is_url(self):
        return any(
            [
                self.value_text.lower().startswith(scheme)
                for scheme in ("http:", "https:", "ftp:")
            ]
        )

    def test_valid(self, pipeline):
        if not self.is_url():
            super(URL, self).test_valid(pipeline)



class ImageSetChannelDescriptor:
    """This class represents the metadata for one image set channel

    An image set has a collection of channels which are either planar
    images or objects. The ImageSetChannelDescriptor describes one
    of these:

    The channel's name

    The channel's type - grayscale image / color image / objects / mask
    or illumination function
    """

    # Channel types
    CT_GRAYSCALE = "Grayscale"
    CT_COLOR = "Color"
    CT_MASK = "Mask"
    CT_OBJECTS = "Objects"
    CT_FUNCTION = "Function"

    def __init__(self, name, channel_type):
        self.name = name
        self.channel_type = channel_type



from ._setting import Setting


class FigureSubscriber(Setting):
    """A setting that subscribes to a figure indicator provider
    """

    def __init__(self, text, value="Do not use", *args, **kwargs):
        super(Setting, self).__init__(text, value, *args, **kwargs)

    def get_choices(self, pipeline):
        choices = []
        for module in pipeline.modules():
            for setting in module.visible_settings():
                if setting.key() == self.key():
                    return choices
            choices.append("%d: %s" % (module.module_num, module.module_name))
        assert False, "Setting not among visible settings in pipeline"



from ._run_exception import RunException


class PostRunException(RunException):
    """An event indicating an uncaught exception during the post_run phase"""

    def event_type(self):
        return "Post run exception"



import os
import sys
import threading

from importlib.util import find_spec


def find_python():
    if hasattr(sys, "frozen"):
        if sys.platform == "darwin":
            app_python = os.path.join(os.path.dirname(os.environ["ARGVZERO"]), "python")
            return app_python
    return sys.executable


def find_worker_env(idx):
    """Construct a command-line environment for the worker

    idx - index of the worker, e.g., 0 for the first, 1 for the second...
    """
    newenv = os.environ.copy()
    cp_install = find_spec("cellprofiler")
    if cp_install:
        root_dir = os.path.abspath(
            os.path.join(os.path.dirname(cp_install.origin), "..")
        )
    else:
        root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))
    added_paths = []
    if "PYTHONPATH" in newenv:
        old_path = newenv["PYTHONPATH"]
        if not any([root_dir == path for path in old_path.split(os.pathsep)]):
            added_paths.append(root_dir)
    else:
        added_paths.append(root_dir)

    if hasattr(sys, "frozen"):
        if sys.platform == "darwin":
            # http://mail.python.org/pipermail/pythonmac-sig/2005-April/013852.html
            added_paths += [p for p in sys.path if isinstance(p, str)]
    if "PYTHONPATH" in newenv:
        added_paths.insert(0, newenv["PYTHONPATH"])
    newenv["PYTHONPATH"] = os.pathsep.join([x for x in added_paths])
    if "CP_JDWP_PORT" in newenv:
        del newenv["CP_JDWP_PORT"]
    if "AW_JDWP_PORT" in newenv:
        port = str(int(newenv["AW_JDWP_PORT"]) + idx)
        newenv["CP_JDWP_PORT"] = port
        del newenv["AW_JDWP_PORT"]
    for key in newenv:
        if isinstance(newenv[key], str):
            newenv[key] = newenv[key]
    return newenv


def find_analysis_worker_source():
    # import here to break circular dependency.
    import cellprofiler_core.worker  # used to get the path to the code

    return os.path.join(
        os.path.dirname(cellprofiler_core.worker.__file__), "__init__.py"
    )


def start_daemon_thread(target=None, args=(), kwargs=None, name=None):
    thread = threading.Thread(target=target, daemon=True, args=args, kwargs=kwargs, name=name)
    thread.start()
    return thread


def close_all_on_exec():
    """Mark every file handle above 2 with CLOEXEC

    We don't want child processes inheret anything
    except for STDIN / STDOUT / STDERR. This should
    make it so in a horribly brute-force way.
    """
    import fcntl

    try:
        maxfd = os.sysconf("SC_OPEN_MAX")
    except:
        maxfd = 256
    for fd in range(3, maxfd):
        try:
            fcntl.fcntl(fd, fcntl.FD_CLOEXEC)
        except:
            pass



from ._setting import Setting
from ._validation_error import ValidationError


class Measurement(Setting):
    """A measurement done on a class of objects (or Experiment or Image)

    A measurement represents a fully-qualified name of a measurement taken
    on an object. Measurements have categories and feature names and
    may or may not have a secondary image (for instance, the image used
    to measure an intensity), secondary object (for instance, the parent
    object when relating two classes of objects or the object name when
    aggregating object measurements over an image) or scale.
    """

    def __init__(self, text, object_fn, value="None", *args, **kwargs):
        """Construct the measurement category subscriber setting

        text - Explanatory text that appears to the side of the setting
        object_fn - a function that returns the measured object when called
        value - the initial value of the setting
        """
        super(Measurement, self).__init__(text, value, *args, **kwargs)
        self.__object_fn = object_fn

    @staticmethod
    def construct_value(category, feature_name, object_or_image_name, scale):
        """Construct a value that might represent a partially complete value"""
        if category is None:
            value = "None"
        elif feature_name is None:
            value = category
        else:
            parts = [category, feature_name]
            if object_or_image_name is not None:
                parts.append(object_or_image_name)
            if not scale is None:
                parts.append(scale)
            value = "_".join(parts)
        return str(value)

    def get_measurement_object(self):
        """Return the primary object for the measurement

        This is either "Image" if an image measurement or the name
        of the objects for per-object measurements. Please pardon the
        confusion with get_object_name which is the secondary object
        name, for instance for a measurement Relate."""
        return self.__object_fn()

    def get_category_choices(self, pipeline, object_name=None):
        """Find the categories of measurements available from the object """
        if object_name is None:
            object_name = self.__object_fn()
        categories = set()
        for module in pipeline.modules():
            if self.key() in [x.key() for x in module.settings()]:
                break
            categories.update(module.get_categories(pipeline, object_name))
        result = list(categories)
        result.sort()
        return result

    def get_category(self, pipeline, object_name=None):
        """Return the currently chosen category"""
        categories = self.get_category_choices(pipeline, object_name)
        for category in categories:
            if self.value.startswith(category + "_") or self.value == category:
                return category
        return None

    def get_feature_name_choices(self, pipeline, object_name=None, category=None):
        """Find the feature name choices available for the chosen category"""
        if object_name is None:
            object_name = self.__object_fn()
        if category is None:
            category = self.get_category(pipeline)
        if category is None:
            return []
        feature_names = set()
        for module in pipeline.modules():
            if self.key() in [x.key() for x in module.settings()]:
                break
            feature_names.update(
                module.get_measurements(pipeline, object_name, category)
            )
        result = list(feature_names)
        result.sort()
        return result

    def get_feature_name(self, pipeline, object_name=None, category=None):
        """Return the currently selected feature name"""
        if category is None:
            category = self.get_category(pipeline, object_name)
        if category is None:
            return None
        feature_names = self.get_feature_name_choices(pipeline, object_name, category)
        for feature_name in feature_names:
            head = "_".join((category, feature_name))
            if self.value.startswith(head + "_") or self.value == head:
                return feature_name
        return None

    def get_image_name_choices(
        self, pipeline, object_name=None, category=None, feature_name=None
    ):
        """Find the secondary image name choices available for a feature

        A measurement can still be valid, even if there are no available
        image name choices. The UI should not offer image name choices
        if no choices are returned.
        """
        if object_name is None:
            object_name = self.__object_fn()
        if category is None:
            category = self.get_category(pipeline, object_name)
        if feature_name is None:
            feature_name = self.get_feature_name(pipeline, object_name, category)
        if category is None or feature_name is None:
            return []
        image_names = set()
        for module in pipeline.modules():
            if self.key() in [x.key() for x in module.settings()]:
                break
            image_names.update(
                module.get_measurement_images(
                    pipeline, object_name, category, feature_name
                )
            )
        result = list(image_names)
        result.sort()
        return result

    def get_image_name(
        self, pipeline, object_name=None, category=None, feature_name=None
    ):
        """Return the currently chosen image name"""
        if object_name is None:
            object_name = self.__object_fn()
        if category is None:
            category = self.get_category(pipeline, object_name)
        if category is None:
            return None
        if feature_name is None:
            feature_name = self.get_feature_name(pipeline, object_name, category)
        if feature_name is None:
            return None
        image_names = self.get_image_name_choices(
            pipeline, object_name, category, feature_name
        )
        # 1st pass - accept only exact match
        # 2nd pass - accept part match.
        # This handles things like "OriginalBlue_Nuclei" vs "OriginalBlue"
        # in MeasureImageIntensity.
        #
        for full_match in True, False:
            for image_name in image_names:
                head = "_".join((category, feature_name, image_name))
                if (not full_match) and self.value.startswith(head + "_"):
                    return image_name
                if self.value == head:
                    return image_name
        return None

    def get_scale_choices(
        self,
        pipeline,
        object_name=None,
        category=None,
        feature_name=None,
        image_name=None,
    ):
        """Return the measured scales for the currently chosen measurement

        The setting may still be valid, even though there are no scale choices.
        In this case, the UI should not offer the user a scale choice.
        """
        if object_name is None:
            object_name = self.__object_fn()
        if category is None:
            category = self.get_category(pipeline, object_name)
        if feature_name is None:
            feature_name = self.get_feature_name(pipeline, object_name, category)
        if image_name is None:
            image_name = self.get_image_name(
                pipeline, object_name, category, feature_name
            )
        if category is None or feature_name is None:
            return []
        scales = set()
        for module in pipeline.modules():
            if self.key() in [x.key() for x in module.settings()]:
                break
            scales.update(
                module.get_measurement_scales(
                    pipeline, object_name, category, feature_name, image_name
                )
            )
        result = [str(scale) for scale in scales]
        result.sort()
        return result

    def get_scale(
        self,
        pipeline,
        object_name=None,
        category=None,
        feature_name=None,
        image_name=None,
    ):
        """Return the currently chosen scale"""
        if object_name is None:
            object_name = self.__object_fn()
        if category is None:
            category = self.get_category(pipeline, object_name)
        if feature_name is None:
            feature_name = self.get_feature_name(pipeline, object_name, category)
        if image_name is None:
            image_name = self.get_image_name(
                pipeline, object_name, category, feature_name
            )
        sub_object_name = self.get_object_name(pipeline)
        if category is None or feature_name is None:
            return None
        if image_name is not None:
            head = "_".join((category, feature_name, image_name))
        elif sub_object_name is not None:
            head = "_".join((category, feature_name, sub_object_name))
        else:
            head = "_".join((category, feature_name))
        for scale in self.get_scale_choices(pipeline):
            if self.value == "_".join((head, scale)):
                return scale
        return None

    def get_object_name_choices(
        self, pipeline, object_name=None, category=None, feature_name=None
    ):
        """Return a list of objects for a particular feature

        Typically these are image features measured on the objects in the image
        """
        if object_name is None:
            object_name = self.__object_fn()
        if category is None:
            category = self.get_category(pipeline, object_name)
        if feature_name is None:
            feature_name = self.get_feature_name(pipeline, object_name, category)
        if any([x is None for x in (object_name, category, feature_name)]):
            return []
        objects = set()
        for module in pipeline.modules():
            if self.key in [x.key() for x in module.settings()]:
                break
            objects.update(
                module.get_measurement_objects(
                    pipeline, object_name, category, feature_name
                )
            )
        result = list(objects)
        result.sort()
        return result

    def get_object_name(
        self, pipeline, object_name=None, category=None, feature_name=None
    ):
        """Return the currently chosen image name"""
        if object_name is None:
            object_name = self.__object_fn()
        if category is None:
            category = self.get_category(pipeline, object_name)
        if category is None:
            return None
        if feature_name is None:
            feature_name = self.get_feature_name(pipeline, object_name, category)
        if feature_name is None:
            return None
        object_names = self.get_object_name_choices(
            pipeline, object_name, category, feature_name
        )
        for object_name in object_names:
            head = "_".join((category, feature_name, object_name))
            if self.value.startswith(head + "_") or self.value == head:
                return object_name
        return None

    def test_valid(self, pipeline):
        obname = self.__object_fn()
        category = self.get_category(pipeline, obname)
        if category is None:
            raise ValidationError(
                "%s has an unavailable measurement category" % self.value, self
            )
        feature_name = self.get_feature_name(pipeline, obname, category)
        if feature_name is None:
            raise ValidationError(
                "%s has an unmeasured feature name" % self.value, self
            )
        #
        # If there are any image names or object names, then there must
        # be a valid image name or object name
        #
        image_name = self.get_image_name(pipeline, obname, category, feature_name)
        image_names = self.get_image_name_choices(
            pipeline, obname, category, feature_name
        )
        sub_object_name = self.get_object_name(pipeline, obname, category, feature_name)
        sub_object_names = self.get_object_name_choices(
            pipeline, obname, category, feature_name
        )
        if len(sub_object_names) > 0 and image_name is None and sub_object_name is None:
            raise ValidationError(
                "%s has an unavailable object name" % self.value, self
            )
        if len(image_names) > 0 and image_name is None and sub_object_name is None:
            raise ValidationError("%s has an unavailable image name" % self.value, self)
        scale_choices = self.get_scale_choices(pipeline, obname, category, feature_name)
        if (
            self.get_scale(pipeline, obname, category, feature_name)
            not in scale_choices
            and len(scale_choices) > 0
        ):
            raise ValidationError("%s has an unavailable scale" % self.value, self)
        for module in pipeline.modules():
            if self.key() in [s.key() for s in module.visible_settings()]:
                break
        if not any(
            [
                column[0] == obname and column[1] == self.value
                for column in pipeline.get_measurement_columns(module)
            ]
        ):
            raise ValidationError(
                "%s is not measured for %s" % (self.value, obname), self
            )



X_AUTOMATIC_EXTRACTION = "Extract from image file headers"
X_MANUAL_EXTRACTION = "Extract from file/folder names"
X_IMPORTED_EXTRACTION = "Import from file"
X_ALL_EXTRACTION_METHODS = [
    X_MANUAL_EXTRACTION,
    X_IMPORTED_EXTRACTION,
]
XM_FILE_NAME = "File name"
XM_FOLDER_NAME = "Folder name"
DTC_TEXT = "Text"
DTC_CHOOSE = "Choose for each"
DTC_ALL = [DTC_TEXT, DTC_CHOOSE]
F_ALL_IMAGES = "All images"
F_FILTERED_IMAGES = "Images matching a rule"
COL_PATH = "Path / URL"
COL_SERIES = "Series"
COL_INDEX = "Frame"
COL_URL = "FileLocation"
DEFAULT_METADATA_TAGS = (COL_URL, COL_SERIES, COL_INDEX)
IDX_EXTRACTION_METHOD_COUNT = 3
IDX_EXTRACTION_METHOD_COUNT_V1 = 1
IDX_EXTRACTION_METHOD_COUNT_V2 = 1
IDX_EXTRACTION_METHOD_COUNT_V3 = 1
IDX_EXTRACTION_METHOD = 4
IDX_EXTRACTION_METHOD_V1 = 2
IDX_EXTRACTION_METHOD_V2 = 2
IDX_EXTRACTION_METHOD_V3 = 2
LEN_EXTRACTION_METHOD_V1 = 8
LEN_EXTRACTION_METHOD = 9
CSV_JOIN_NAME = "CSV Metadata"
IPD_JOIN_NAME = "Image Metadata"



import os
import pathlib

import click

import cellprofiler_core.commands

CONTEXT_SETTINGS = dict(auto_envvar_prefix="COMPLEX")


class Command(click.MultiCommand):
    def get_command(self, context, name):
        try:
            name = f"cellprofiler_core.commands._{name}_command"

            imported_module = __import__(name, None, None, ["command"])
        except ImportError:
            return

        return imported_module.command

    def list_commands(self, context):
        command_names = []

        commands_pathname = cellprofiler_core.commands.__file__

        commands_directory = pathlib.Path(commands_pathname).parent

        for filename in os.listdir(commands_directory):
            if filename.endswith("_command.py") and filename.startswith("_"):
                command_name = filename[1:-11]

                command_names += [command_name]

        command_names.sort()

        return command_names


class Environment:
    def __init__(self):
        pass


pass_environment = click.make_pass_decorator(Environment, ensure=True)


@click.command(cls=Command, context_settings=CONTEXT_SETTINGS)
@pass_environment
def main(context):
    pass


if __name__ == "__main__":
    main({})
